// Address.h
//
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2013 Mike McCauley
// $Id: Address.h,v 1.2 2013/07/14 07:15:46 mikem Exp mikem $


#ifndef MAP27_ADDRESS_H
#define MAP27_ADDRESS_H

#include <stdint.h>

// Max possible address component length
#define MAP27_ADDRESS_MAX_ALEN            15

// Some address and ident definitions
// Address types ADESC
#define MAP27_ADDRESS_ADESC_NO_ADDRESS          0x00
#define MAP27_ADDRESS_ADESC_SAMIS_BCD           0x01
#define MAP27_ADDRESS_ADESC_SPARE               0x02
#define MAP27_ADDRESS_ADESC_MPT1327_PABX_BINARY 0x03
#define MAP27_ADDRESS_ADESC_MPT1327_PABX        0x04
#define MAP27_ADDRESS_ADESC_MPT1327             0x05
#define MAP27_ADDRESS_ADESC_MPT1327_SAMIS       0x06
#define MAP27_ADDRESS_ADESC_MMI_DIALLED_NUMBER  0x07

// Well known prefixes and idents
#define MAP27_ADDRESS_DUMMYP              0x00
#define MAP27_ADDRESS_DUMMYI              0x00
#define MAP27_ADDRESS_PSTNGI              8101
#define MAP27_ADDRESS_PABXI               8102
#define MAP27_ADDRESS_DNI                 8103

#define MAP27_ADDRESS_PSTNI(j)            (8120+j)
#define MAP27_ADDRESS_REGI                8185
#define MAP27_ADDRESS_INCI                8186
#define MAP27_ADDRESS_DIVERTI             8187
#define MAP27_ADDRESS_SDMI                8188
#define MAP27_ADDRESS_IPFIXI              8189
#define MAP27_ADDRESS_TSCI                8190
#define MAP27_ADDRESS_ALLI                0000

/////////////////////////////////////////////////////////////////////
/// \class Map27Address Address.h <Map27/Address.h>
/// \brief Class for handling Map27 Prefix/Ident/Address
///
/// This class provides methods to hold and encapsulate Map27 Prefix, Ident and optional extra Address
/// representations, as well as functions for encoding and decoding them into on-the-wire
/// message formats.
///
/// A Map27Address always has a MPT1327 Prefix and Ident. It can also have an optional 
/// Address which can be used for specifying PABX number etc.
///
/// Some Map27 messages only contain the Prefix and Ident and no Address. 
/// We call these 'Basic' and there are separate
/// methods for encoding and decoding them, which ignore the Address component.
class Map27Address
{
public:

    /// Constructor. Prefix defaults to MAP27_ADDRESS_DUMMYP.
    /// Ident defaults to MAP27_ADDRESS_DUMMYI. Address defaults to MAP27_ADDRESS_NO_ADDRESS.
    Map27Address();

    /// Set the MPT1327 prefix
    /// \param[in] prefix The Prefix to set
    void     setPrefix(uint8_t prefix = MAP27_ADDRESS_DUMMYP);

    /// Get the current MPT1327 prefix
    /// \return The current MPT1327 prefix
    uint8_t  prefix() const;

    /// Set the MPT1327 Unit Ident (UI)
    /// \param[in] ident The Ident to set
    void     setIdent(uint16_t ident = MAP27_ADDRESS_DUMMYI);

    /// Get the current MPT1327 Unit Ident (UI)
    /// return The current MPT1327 Unit Ident (UI)
    uint16_t ident() const;

    /// Set the current optional Address
    /// An address includes the addres type (adesc), an a number of bytes of address information
    /// Map27Address takes care of the encoding of the Address in message rypes that require it
    /// \param[in] adesc Address type. One of MAP27_ADDRESS_ADESC_*
    /// \param[in] address Array of encoded address bytes.
    /// \param[in] alen Number of bytes in address
    void     setAddress(uint8_t adesc = MAP27_ADDRESS_ADESC_NO_ADDRESS, uint8_t* address = 0, uint8_t alen = 0);

    /// Get the current Address type
    /// \return the current Addres type, one of MAP27_ADDRESS_ADESC_*
    uint8_t  adesc() const;

    /// Get the current address length
    /// \return The current Address length in bytes
    uint8_t  alen() const;

    /// Get a pointer to the current address
    /// Caution, this is internal storage and may not be overwritten.
    /// The number of valid address bytes can be got with alen()
    /// \return Pointer to the internal storage of the current Address
    const uint8_t* address() const;

    /// Decode Prefix and Ident (but not Address) from on-the-wire format.
    /// Sets the internal Prefix and Ident. Sets the address type to MAP27_ADDRESS_DESC_NO_ADDRESS
    /// and the Address length to 0
    /// \param[in] buf Pointer to the buffer containing the on-the-wire encoded Prefix and Ident
    /// \param[in] bufLen Length of buf in bytes
    /// \return The number of bytes decoded. Normally 3. Returns 0 on failure (ie buffer not long 
    /// enough to contain a Prefix and Ident).
    uint8_t  decodeBasic(uint8_t* buf, uint16_t bufLen);

    /// Decode Prefix, Ident amnd Address from on-the-wire format.
    /// \param[in] buf Pointer to the buffer containing the on-the-wire encoded Prefix, Ident and Address
    /// \param[in] bufLen Length of buf in bytes
    /// \return The number of bytes decoded. Normally 4 or more. Returns 0 on failure (ie buffer not long 
    /// enough to contain a valid Address).
    uint8_t  decode(uint8_t* buf, uint16_t bufLen);

    /// Encode Prefix and Ident (but not Address) in on-the-wire format to a buffer
    /// \param[in] buf The buffer where the encoded data is to be written
    /// \param[in] bufLen Length of available spacec in buf in bytes
    /// \return The number of bytes written to buf. 0 on failure (such as insufficient space in buf)
    uint8_t  encodeBasic(uint8_t* buf, uint16_t bufLen) const;

    /// Encode Prefix, Ident and Address in on-the-wire format to a buffer
    /// \param[in] buf The buffer where the encoded data is to be written
    /// \param[in] bufLen Length of available spacec in buf in bytes
    /// \return The number of bytes written to buf. 0 on failure (such as insufficient space in buf)
    uint8_t  encode(uint8_t* buf, uint16_t bufLen) const;

    /// Get the current prefix in MPT1343 form
    /// \return The current prefix in MPT1343 form
    uint16_t mpt1343_prefix() const;

    /// Convert the given Individual Base Ident to Fleet Individual Number
    /// \param[in] ibi MPT1327 Individual Base Ident
    /// \return MPT1343 Fleet Individual Number
    uint16_t mpt1343_fin(uint16_t ibi) const;

    /// Convert the current MPT1327 Ident and IBI to MPT1343 Unit Number 3 Digit (UN3D)
    /// \param[in] ibi MPT1327 Individual Base Ident
    /// \return MPT1343 Unit Number 3 Digit 
    uint16_t mpt1343_un3d(uint16_t ibi) const;

    /// Convert the current MPT1327 Ident and IBI to MPT1343 Unit Number 2 Digit (UN2D)
    /// \param[in] ibi MPT1327 Individual Base Ident
    /// \return MPT1343 Unit Number 2 Digit 
    uint16_t mpt1343_un2d(uint16_t ibi) const;

    /// Formats the address as a string
    /// and returns a pointer to temporary storage containing the string
    /// \return Pointer to formatted string in private storage
    const char* asString();

private:
    uint8_t     _prefix;
    uint16_t    _ident;
    uint8_t     _adesc;
    uint8_t     _alen;
    uint8_t     _address[MAP27_ADDRESS_MAX_ALEN];
};

#endif
