// ApplicationLayer.h
//
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2013 Mike McCauley
// $Id: ApplicationLayer.h,v 1.1 2013/07/14 06:32:28 mikem Exp $


#ifndef MAP27_APPLICATIONLAYER_H
#define MAP27_APPLICATIONLAYER_H

#include <stdint.h>

// Error types
#define MAP27_APPLICATION_ERROR_NO_ERROR               0
#define MAP27_APPLICATION_ERROR_REQUEST_TIMEOUT        1
#define MAP27_APPLICATION_ERROR_OPERATION_ABORTED      2
#define MAP27_APPLICATION_ERROR_INVALID_STATE          3

class Map27Address;
class Map27NetworkLayer;

/////////////////////////////////////////////////////////////////////
/// \class Map27ApplicationLayer ApplicationLayer.h <Map27/ApplicationLayer.h>
/// \brief Superclass for a generic Map27 application
///
/// This class is intended as a generic superclass, to be used as the base for constructing
/// a specific Map27 application. Application developers are expected to create a specific subclass
/// for their application, instantiate it and connect it to a Map27NetworkLayer instance.
class Map27ApplicationLayer
{
public:
    /// Constructor
    Map27ApplicationLayer();

    /// Sets the pointer to the NetworkLayer to use
    /// to communicate with the Map27 radio.
    /// \param[in] networkLayer Pointer to the Network Layer
    void setNetworkLayer(Map27NetworkLayer* networkLayer);

    /// Get the current Network Layer pointer.
    /// \return The current Network Layer pointer, or NULL if previously set.
    Map27NetworkLayer* networkLayer();

    /// Powers on the application and lower layers
    /// Calls power_on() in the Network Layer.
    /// You may consider overriding this to provide Application specific behaviour
    void power_on();

    /// Polls the Network Layer and any of its lower layers for activity.
    /// In Unix/Linux and Arduino environments, you are required to call this at least as 
    /// frequently as octets might be received from the radio. At least every 10msec would 
    /// be appropriate for a 9600 baud connection
    void poll();

    /// \name Inputs from Network Layer
    /// Override these in your subclass to get control when certain events happen.
    /// The default implementations do nothing but log the event at DEBUG level.

    /// Called by Network Layer
    /// when an error is being reported to the Application Layer.
    /// \param[in] type The error type. One of MAP27_APPLICATION_ERROR_*
    virtual void error(uint8_t type);

    /// Called by Network Layer
    /// when a RECEIVE STATUS (5.2.2.1.2) is received.
    /// \param[in] address The Address of the calling party
    /// \param[in] status The status, one of MAP27_NETWORK_MESSAGE_STATUS_*
    virtual void receiveStatus(Map27Address& address, uint8_t status);

    /// Called by Network Layer
    /// when a STATUS ACK (5.2.2.1.3) message is received.
    /// \param[in] address The Address of the calling party
    /// \param[in] cause Reason for sending this message. One of MAP27_NETWORK_MESSAGE_STATUS_CAUSE_*
    virtual void receiveStatusAck(Map27Address& address, uint8_t cause);

    /// Called by Network Layer
    /// when a CLEARED (5.2.2.6.2) message is received.
    /// \param[in] address The Address of the calling, called or connected party
    /// \param[in] cause Cancellation or disconnection reason. One of MAP27_NETWORK_MESSAGE_CLEARED_CAUSE_*
    virtual void receiveCleared(Map27Address& address, uint8_t cause);

    /// Called by Network Layer
    /// when a RECEIVE SST (5.2.2.2.2) message is received.
    /// \param[in] address The Address of the calling party
    /// \param[in] buf Pointer to the octets of the received SST message, starting with the CODING/NLB octet
    /// \param[in] bufLen The number of octets in thje received message, including the CODING/NLB octet
    virtual void receiveSST(Map27Address& address, uint8_t* buf, uint8_t bufLen);

    /// Called by Network Layer
    /// when a SST ACK (5.2.2.2.3) message is received.
    /// \param[in] address The Address of the called party
    /// \param[in] cause Reason for sending this message. One of MAP27_NETWORK_MESSAGE_SST_ACK_*
    virtual void receiveSSTAck(Map27Address& address, uint8_t cause);

    /// Called by Network Layer
    /// when a RECEIVE MST (5.2.2.3.2) message is received.
    /// \param[in] address The Address of the calling party
    /// \param[in] buf Pointer to the octets of the received MST message, starting with the CODING/NLB octet
    /// \param[in] bufLen The number of octets in thje received message, including the CODING/NLB octet
    virtual void receiveMST(Map27Address& address, uint8_t* buf, uint8_t bufLen);

    /// Called by Network Layer
    /// when a MST ACK (5.2.2.3.3) message is received.
    /// \param[in] address The Address of the called party
    /// \param[in] cause Reason for sending this message. One of MAP27_NETWORK_MESSAGE_SST_ACK_*
    /// or MAP27_NETWORK_MESSAGE_MST_ACK_*
    virtual void receiveMSTAck(Map27Address& address, uint8_t cause);

    /// Called by Network Layer
    /// when a SETUP PROGRESS (5.2.2.4.3) message is received.
    /// \param[in] address The Address of the called party
    /// \param[in] cause Reason for sending this message. One of MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_*
    virtual void receiveSetupProgress(Map27Address& address, uint8_t cause);

    /// Called by Network Layer
    /// when an INCOMING VOICE CALL / INCOMING MODEM CALL (5.2.2.4.4) message is received.
    /// \param[in] address The Address of the calling party
    /// \param[in] details Call details. One of MAP27_NETWORK_MESSAGE_INCOMING_VOICE_MODEM_DETAILS_*
    /// \param[in] buf Parameter bits. Up to 44 bits in 5 and a half octets, only with customised services.
    /// \param[in] bufLen Number of octets in buf.
    virtual void receiveIncomingVoiceModem(Map27Address& address, uint8_t details, uint8_t* buf, uint8_t bufLen);

    /// Called by Network Layer
    /// when a RECEIVE PROGRESS (5.2.2.4.6) message is received.
    /// \param[in] address The Address of the calling party
    /// \param[in] cause Reason for sending this message. One of MAP27_NETWORK_MESSAGE_RECEIVE_PRGRESS_CAUSE_*
    virtual void receiveProgress(Map27Address& address, uint8_t cause);

    /// Called by Network Layer
    /// when a RECEIVE MODEM DATA (5.2.2.5.2) message is received.
    /// \param[in] buf Pointer to start of data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveModemData(uint8_t* buf, uint8_t bufLen);

    /// Called by Network Layer
    /// when a RADIO PERSONALITY (5.2.2.8.2) message is received.
    /// \param[in] address Radio unit address
    /// \param[in] manufacturers_code Manufacturers code
    /// \param[in] model Radio model number
    /// \param[in] serial Radio serial number
    /// \param[in] facilities_a List of supported facilities. Mask of MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_FACILITIESA_*
    /// \param[in] facilities_b Reserved
    /// \param[in] facilities_c Reserved
    /// \param[in] codings Supported codings. Mask of MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_CODINGS_
    virtual void receiveRadioPersonality(Map27Address& address, uint8_t manufacturers_code, uint8_t model, 
					 uint32_t serial, uint8_t facilities_a, uint8_t facilities_b, 
					 uint8_t facilities_c, uint8_t codings);

    /// Called by Network Layer
    /// when a PROTOCOL INFO (5.2.2.8.9) message is received.
    /// \param[in] reason Reason to send this message. One of MAP27_NETWORK_MESSAGE_PROTOCOL_INFO_
    virtual void receiveProtocolInfo(uint8_t reason);

    /// Called by Network Layer
    /// when a NUMBERING INFORMATION (5.2.2.8.3) message is received.
    /// \param[in] address Radio unit address
    /// \param[in] individual_base individual number base ident (IBI)
    /// \param[in] individual_base_3digit IBI is 3-digit
    /// \param[in] highest_ident Highest indviual ident
    /// \param[in] group_base Group number base ident
    /// \param[in] group_base_3digit Group number base ident is 3-digit
    /// \param[in] highest_group Highest group ident
    virtual void receiveNumberingInfo(Map27Address& address, 
				      uint16_t individual_base, bool individual_base_3digit, uint16_t highest_ident, 
				      uint16_t group_base, bool group_base_3digit, uint16_t highest_group);

    /// Called by Network Layer
    /// whena RADIO SETTINGS (5.2.2.8.8) message is received.
    /// \param[in] controls_a Activated facilities, Mask of MAP27_NETWORK_MESSAGE_RADIO_MANAGEMENT_CONTROLS_*
    /// \param[in] controls_b Reserved
    /// \param[in] controls_c Reserved
    /// \param[in] codings Activated coding type. Mask of MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_CODINGS_*
    virtual void receiveRadioSettings(uint8_t controls_a, uint8_t controls_b, uint8_t controls_c, uint8_t codings);

    /// Called by Network Layer
    /// when a OPERATING CONDITION (5.2.2.8.5) message is received.
    /// \param[in] conditions Operating conditions. Mask of MAP27_NETWORK_MESSAGE_OPERATING_CONDITION_CONDITION_*
    /// \param[in] field_strength Received field strength. One of MAP27_NETWORK_MESSAGE_OPERATING_CONDITION_FIELD_STRENGTH_*
    /// \param[in] max_call_duration Maximum call duration time. One of MAP27_NETWORK_MESSAGE_OPERATING_CONDITION_CALL_DURATION_* and others.
    virtual void receiveOperatingCondition(uint8_t conditions, uint8_t field_strength, uint8_t max_call_duration);

    /// Called by Network Layer
    /// when a NETWORK INFORMATION (5.2.2.8.6) message is received.
    /// \param[in] channel Radio channel number
    /// \param[in] sys_ident System identity code
    virtual void receiveNetworkInformation(uint16_t channel, uint16_t sys_ident);

    /// Called by Network Layer
    /// when a RADIO TEST RESULT (5.2.2.8.13) message is received.
    /// \param[in] buf Pointer to test results data.
    /// \param[in] bufLen Number of bytes in buf.
    virtual void receiveRadioTestResult(uint8_t* buf, uint8_t bufLen);

    /// \name Network layer events
    /// These functions register other Network Layer events

    /// Called by the Network Layer when the
    /// RS232 link to the Radio is established or re-established
    virtual void link_ready();

    /// Called by the Network Layer when the
    /// RS232 link to the Radio is lost
    virtual void link_failure();

    /// Called by the Network Layer when 
    /// a mesage is successfully queued for delivery to the radio
    virtual void packet_accepted();

    /// Called by the Network Layer when 
    /// a mesage is rejected for delivery to the radio
    virtual void packet_rejected();

private:
    Map27NetworkLayer* _networkLayer;  // The current Network Layer
};

#endif
