// Data.cpp
//
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2013 Mike McCauley
// $Id: DataLayer.cpp,v 1.10 2013/07/14 06:42:42 mikem Exp $

// For memcpy:
#include <string.h>

#include "Map27.h"
#include "DataLayer.h"
#include "PhysicalLayer.h"
#include "NetworkLayer.h"
#include "Log.h"

// Timer callback glue
void Map27DataLayer::link_establishment_timer_callback_glue(void* target)
{
    ((Map27DataLayer*)target)->link_establishment_timeout();
}

void Map27DataLayer::retry_timer_callback_glue(void* target)
{
    ((Map27DataLayer*)target)->retry_timeout();
}

void Map27DataLayer::acknowledgement_timer_callback_glue(void* target)
{
    ((Map27DataLayer*)target)->acknowledgement_timeout();
}

void Map27DataLayer::activity_timer_callback_glue(void* target)
{
    ((Map27DataLayer*)target)->activity_timeout();
}

void Map27DataLayer::link_failure_detection_timer_callback_glue(void* target)
{
    ((Map27DataLayer*)target)->link_failure_detection_timeout();
}

Map27DataLayer::Map27DataLayer()
{
    _physicalLayer = NULL;
    _networkLayer = NULL;

    _baud = MAP27_DEFAULT_BAUD;

    _link_state = LINK_STATE_RESET_WAIT;

    // Set up timer callbacks
    _link_establishment_timer.setCallback(link_establishment_timer_callback_glue, this);
    _retry_timer.setCallback(retry_timer_callback_glue, this);
    _acknowledgement_timer.setCallback(acknowledgement_timer_callback_glue, this);
    _activity_timer.setCallback(activity_timer_callback_glue, this);
    _link_failure_detection_timer.setCallback(link_failure_detection_timer_callback_glue, this);

    // Clear transmit buffers
    _txQueueLength = 0;
    _txBottom = 0;
    _txNext = 0;
}

void Map27DataLayer::poll()
{
    if (_physicalLayer)
	_physicalLayer->poll();
    _link_establishment_timer.poll();
    _retry_timer.poll();
    _acknowledgement_timer.poll();
    _activity_timer.poll();
    _link_failure_detection_timer.poll();
}

void Map27DataLayer::setPhysicalLayer(Map27PhysicalLayer* physicalLayer)
{
    _physicalLayer = physicalLayer;
}

Map27PhysicalLayer* Map27DataLayer::physicalLayer()
{
    return _physicalLayer;
}

void Map27DataLayer::setNetworkLayer(Map27NetworkLayer* networkLayer)
{
    _networkLayer = networkLayer;
}

Map27NetworkLayer* Map27DataLayer::networkLayer()
{
    return _networkLayer;
}

void Map27DataLayer::setBaud(uint32_t baud)
{
    _baud = baud;
}

uint32_t Map27DataLayer::baud()
{
    return _baud;
}

void Map27DataLayer::power_on()
{
    if (   _link_state == LINK_STATE_LINK_WAIT
	|| _link_state == LINK_STATE_READY)
	output_link_failure();
    reset();
}

void Map27DataLayer::network_layer_reset()
{
    reset();
}

void Map27DataLayer::credit_value(uint8_t r_k)
{
    DEBUG_PRINTF("credit_value %d\n", r_k);
    // From network layer
    if (_link_state == LINK_STATE_READY)
    {
	update_receive_credit(r_k);
	if (receive_credit_available())
	{
	    _acknowledgement_timer.stop();
	    output_link_acknowledge();
	    _activity_timer.start(MAP27_ACTIVITY_TIMEOUT);
	}
    }
}

void Map27DataLayer::reset()
{
    initialise_rs232_port();
    resetLink();
}

void Map27DataLayer::resetLink()
{
    DEBUG_PRINTF("RESET LINK\n");
    // Clear internal buffers
    _txQueueLength = 0;
    _txBottom = 0;
    _txNext = 0;

    // Do everything required to go to RESET_WAIT state with prejudice
    maximise_link_parameters();
    _link_establishment_timer.stop();
    _retry_timer.stop();
    _acknowledgement_timer.stop();
    _activity_timer.stop();
    _link_failure_detection_timer.stop();
    output_link_request();
    _link_establishment_timer.start(MAP27_ESTABLISHMENT_TIMEOUT);
    _link_state = LINK_STATE_RESET_WAIT;
}

void Map27DataLayer::link_request(uint8_t n1, uint8_t k, uint8_t version)
{
    DEBUG_PRINTF("RX LR linkRequest %d: %d %d %d\n", _link_state, n1, k, version);
    if (_link_state == LINK_STATE_RESET_WAIT)
    {
	adjust_link_parameters(n1, k, version);
	output_link_request();
	_link_establishment_timer.start(MAP27_ESTABLISHMENT_TIMEOUT);
	_link_state = LINK_STATE_LINK_WAIT;
    }
    else if (_link_state == LINK_STATE_LINK_WAIT)
    {
	adjust_link_parameters(n1, k, version);
	if (received_parameters_acceptable(n1, k, version))
	{
	    initialise_variables();
	    _link_establishment_timer.stop();
	    output_link_acknowledge();
	    _activity_timer.start(MAP27_ACTIVITY_TIMEOUT);
	    _link_failure_detection_timer.start(MAP27_LINK_FAILURE_DETECTION_TIMEOUT);
	    _link_state = LINK_STATE_READY;
	    output_link_ready();
	}
	else
	{
	    output_link_request();
	    _link_establishment_timer.start(MAP27_ESTABLISHMENT_TIMEOUT);
	}
    }
    else if (_link_state == LINK_STATE_READY)
    {
	adjust_link_parameters(n1, k, version);
	output_link_failure();
	_link_establishment_timer.stop();
	_retry_timer.stop();
	_acknowledgement_timer.stop();
	_activity_timer.stop();
	_link_failure_detection_timer.stop();
	output_link_request();
	_link_establishment_timer.start(MAP27_ESTABLISHMENT_TIMEOUT);
	_link_state = LINK_STATE_LINK_WAIT;
    }
}

void Map27DataLayer::link_acknowledge(uint8_t n_r, uint8_t n_k)
{
    DEBUG_PRINTF("RX LA linkAcknowledge  %d: %d %d\n", _link_state, n_r, n_k);

    _second_last_n_r = _last_n_r;
    _last_n_r = n_r;
    if (_link_state == LINK_STATE_RESET_WAIT)
    {
	output_link_request();
	_link_establishment_timer.start(MAP27_ESTABLISHMENT_TIMEOUT);
    }
    else if (_link_state == LINK_STATE_LINK_WAIT)
    {
	initialise_variables();
	record_send_credit(n_r, n_k);
	_link_establishment_timer.stop();
	output_link_acknowledge();
	_activity_timer.start(MAP27_ACTIVITY_TIMEOUT);
	_link_failure_detection_timer.start(MAP27_LINK_FAILURE_DETECTION_TIMEOUT);
	_link_state = LINK_STATE_READY;
	output_link_ready();
    }
    else if (_link_state == LINK_STATE_READY)
    {
	if (acknowledgement_inside_window(n_r))
	{
	    DEBUG_PRINTF("ack inside window\n");
	    // The link is still working
	    _link_failure_detection_timer.start(MAP27_LINK_FAILURE_DETECTION_TIMEOUT);
	    delete_acknowledged_packets(n_r);
	    // moved:
	    store_acknowledged_rx_sequence_number(n_r);
	    record_send_credit(n_r, n_k);
	    if (all_transmitted_packets_acknowledged(n_r))
	    {
		DEBUG_PRINTF("all_transmitted_packets_acknowledged\n");
		_retry_timer.stop();
		set_retransmission_counter(MAP27_MAX_RETRANSMISSIONS);
	    }
	    else
	    {
		DEBUG_PRINTF("not all_transmitted_packets_acknowledged\n");
		if (repeated_link_acknowledge(n_r))
		{
		    DEBUG_PRINTF("repeated_link_acknowledge\n");
		    // Request to send some packets again
		    decrement_retransmission_counter();
		    if (retransmission_count_zero())
		    {
			DEBUG_PRINTF("retransmission_count_zero\n");
			// Exhausted retrans: bad link?
			resetLink();
			output_link_failure();
		    }
		    else
		    {
			DEBUG_PRINTF("retransmit\n");
			// Retransmit
			rewind_packet_number(n_r);
			output_link_transfer();
			_retry_timer.start(calc_retry_timeout());
		    }
		}
	    }
	}
	else
	{
	    // Acknowledgement is outside window
	    DEBUG_PRINTF("ack outside window\n");
	    resetLink();
	    output_link_failure();
	}
    }
}

void Map27DataLayer::link_transfer(uint8_t n_s, uint8_t ar, uint8_t* buf, uint16_t bufLen)
{
    DEBUG_PRINTF("RX LT linkTransfer  %d: %d %d %d\n", _link_state, n_s, ar, bufLen);
    DEBUG_DUMP("linkTransfer", buf, bufLen);

    if (   _link_state == LINK_STATE_RESET_WAIT
	   || _link_state == LINK_STATE_LINK_WAIT)
    {
	output_link_request();
	_link_establishment_timer.start(MAP27_ESTABLISHMENT_TIMEOUT);
    }
    else if (_link_state == LINK_STATE_READY)
    {
	// The link is working
	_link_failure_detection_timer.start(MAP27_LINK_FAILURE_DETECTION_TIMEOUT);
	if (packet_outside_window(n_s))
	{
	    DEBUG_PRINTF("outside window\n");
	    // Data link error
	    resetLink();
	    output_link_failure();
	}
	else
	{
	    if (packet_out_of_sequence(n_s))
	    {
		DEBUG_PRINTF("out of sequence\n");
		_acknowledgement_timer.stop();
		output_link_acknowledge();
		_activity_timer.start(MAP27_ACTIVITY_TIMEOUT);
	    }
	    else
	    {
		DEBUG_PRINTF("in sequence\n");
		if (receive_credit_available())
		{
		    DEBUG_PRINTF("credit availble\n");
		    increment_receive_state();
		    decrement_receive_credit();
		    if (immediate_reply_requested(ar))
		    {
			DEBUG_PRINTF("immedaite reply requested\n");
			store_receive_credit();
			_acknowledgement_timer.stop();
			output_link_acknowledge();
			_activity_timer.start(MAP27_ACTIVITY_TIMEOUT);
		    }
		    else
		    {
			DEBUG_PRINTF("not immedaite reply requested\n");
			store_receive_credit();
			_acknowledgement_timer.start(calc_acknowledgement_timeout());
		    }
		    output_network_layer_packet(buf, bufLen);
		}
		else
		{
		    DEBUG_PRINTF("no credit availble\n");
		    store_receive_credit();
		    _acknowledgement_timer.stop();
		    output_link_acknowledge();
		    _activity_timer.start(MAP27_ACTIVITY_TIMEOUT);
		}
	    }
	}
    }
}

// Timer Events
void Map27DataLayer::link_establishment_timeout()
{
    DEBUG_PRINTF("link_establishment_timeout %d\n", _link_state);
    if (   _link_state == LINK_STATE_RESET_WAIT
	   || _link_state == LINK_STATE_LINK_WAIT)
    {
	output_link_request();
	_link_establishment_timer.start();
	_link_state = LINK_STATE_RESET_WAIT;
    }
}

void Map27DataLayer::retry_timeout()
{
    if (_link_state == LINK_STATE_READY)
    {
	decrement_retransmission_counter();
	if (retransmission_count_zero())
	{
	    resetLink();
	    output_link_failure();
	}
	else
	{
	    rewind_packet_number(_stored_acknowledged_rx_sequence_number);
	    output_link_transfer();
	    _retry_timer.start(calc_retry_timeout());
	}
    }
}

void Map27DataLayer::acknowledgement_timeout()
{
    DEBUG_PRINTF("acknowledgement_timeout %d\n", _link_state);
    if (_link_state == LINK_STATE_READY)
    {
	output_link_acknowledge();
	_activity_timer.start(MAP27_ACTIVITY_TIMEOUT);
    }
}

void Map27DataLayer::activity_timeout()
{
    DEBUG_PRINTF("activity_timeout %d\n", _link_state);
    if (_link_state == LINK_STATE_READY)
    {
	_acknowledgement_timer.stop();
	output_link_acknowledge();
	_activity_timer.start(MAP27_ACTIVITY_TIMEOUT);
    }
}

void Map27DataLayer::link_failure_detection_timeout()
{
    DEBUG_PRINTF("link_failure_detection_timeout %d\n", _link_state);
    if (_link_state == LINK_STATE_READY)
    {
	resetLink();
	output_link_failure();
    }
}

bool Map27DataLayer::network_layer_packet(uint8_t* buf, uint16_t bufLen)
{
    DEBUG_PRINTF("network_layer_packet\n");
    if (   _link_state == LINK_STATE_RESET_WAIT
	|| _link_state == LINK_STATE_LINK_WAIT)
    {
	output_packet_rejected();
	return false;
    }
    else if (_link_state == LINK_STATE_READY)
    {
	if (send_credit_available())
	{
	    DEBUG_PRINTF("send_credit_available\n");
	    output_packet_accepted();
	    store_packet(buf, bufLen);
	    set_retransmission_counter(MAP27_MAX_RETRANSMISSIONS);
	    output_link_transfer();
	    increment_send_state();
	    decrement_send_credit();
	    _retry_timer.start(calc_retry_timeout());
	    return true;
	}
	else
	{
	    DEBUG_PRINTF("no send_credit_available\n");
	    output_packet_rejected();
	    return false;
	}
    }
}

// TASKS
void Map27DataLayer::adjust_link_parameters(uint8_t n1, uint8_t k, uint8_t version)
{
    _maximum_length = min(n1, MAP27_N1);
    _window_size = min(k, MAP27_WINDOW_SIZE);
    _version = min(version, MAP27_PROTOCOL_VERSION);
}

void Map27DataLayer::decrement_retransmission_counter()
{
    _retransmission_count--;
}

void Map27DataLayer::decrement_send_credit()
{
    _send_credit--;
    DEBUG_PRINTF("decrement_send_credit, now %d\n", _send_credit);
}

void Map27DataLayer::delete_acknowledged_packets(uint8_t n_r)
{
    DEBUG_PRINTF("delete_acknowledged_packets %d %d\n", n_r, _stored_acknowledged_rx_sequence_number);
    // Removes acknowledged messages from the queue of network layer messages awaiting
    // transmission by deleting all messages with 'stored_packet_number' less than the value of
    // 'rx_sequence_number' (N(R)) in the last received link_acknowledgement message.
    uint8_t numToRemove = n_r - _stored_acknowledged_rx_sequence_number;
    DEBUG_PRINTF("delete_acknowledged_packets removes %d\n", numToRemove);
    _txBottom += numToRemove;
    _txQueueLength -= numToRemove;
}

void Map27DataLayer::increment_send_state()
{
    _send_state++;
    DEBUG_PRINTF("_send_state now %d\n", _send_state);
}

void Map27DataLayer::initialise_variables()
{
    _send_state = 1;
    _receive_state = 1;
    _receive_credit = _window_size;
    _send_credit = 0;
    _stored_acknowledged_rx_sequence_number = 1;
    if (_physicalLayer)
	_physicalLayer->clearRxBuf();
    _retransmission_count = MAP27_MAX_RETRANSMISSIONS; // N2

    _last_n_r = 1;
    _second_last_n_r = 1;
}

void Map27DataLayer::initialise_rs232_port()
{
    if (_physicalLayer)
    {
	_physicalLayer->configure_rs232_port(_baud);
	_physicalLayer->reset();
    }
}

void Map27DataLayer::maximise_link_parameters()
{
    _maximum_length = MAP27_N1;
    _window_size = MAP27_WINDOW_SIZE;
    _version = MAP27_PROTOCOL_VERSION;
}

void Map27DataLayer::record_send_credit(uint8_t n_r, uint8_t n_k)
{
    DEBUG_PRINTF("record_send_credit %d %d %d\n", _send_state, n_r, n_k);
    _send_credit = n_k - (_send_state - n_r);
    DEBUG_PRINTF("record_send_credit now %d\n",  _send_credit);
}

void Map27DataLayer::rewind_packet_number(uint8_t n_r)
{
    DEBUG_PRINTF("rewind_packet_number %d %d\n", n_r, _send_state);
    // Set 'send_state', V(S), to the value of the rx_sequence_number parameter (N(R)) of the last
    // received link_acknowledgement message, and set tx_message_pointer to the corresponding
    // message in the tx_messages_buffer, so that it will be the next message to be transmitted.
    _txNext = (_txNext - (_send_state - n_r) + _window_size) % _window_size;
    _send_state = n_r;
    DEBUG_PRINTF("_send_state now %d\n", _send_state);
}

void Map27DataLayer::set_retransmission_counter(uint8_t c1)
{
    _retransmission_count = c1;
}

void Map27DataLayer::store_acknowledged_rx_sequence_number(uint8_t n_r)
{
    _stored_acknowledged_rx_sequence_number = n_r;
}

void Map27DataLayer::store_packet(uint8_t* buf, uint16_t bufLen)
{
    // Stores most recently received packet from network layer, and marks it with a
    // 'stored_packet_number' equal to the current value of the variable 'transmit_buffer_top' and
    // increments the variable 'transmit_buffer_top'.
    DEBUG_PRINTF("store_packet q length: %d\n", _txQueueLength);
    DEBUG_DUMP("store_packet", buf, bufLen);
    if (_txQueueLength < MAP27_WINDOW_SIZE)
    {
	// Room in queue
	DEBUG_PRINTF("store_packet has room\n");
	uint8_t i = (_txBottom + _txQueueLength) % _window_size;
	memcpy(_txQueueBuf[i], buf, bufLen);
	_txQueueBuflen[i] = bufLen;
	_txQueueLength++;
    }
}

void Map27DataLayer::store_receive_credit()
{
    _stored_receive_credit = _receive_credit;
}

void Map27DataLayer::update_receive_credit(uint8_t r_k)
{
    _receive_credit = r_k;
}

void Map27DataLayer::increment_receive_state()
{
    _receive_state++;
}

void Map27DataLayer::decrement_receive_credit()
{
    _receive_credit--;
}

// Returns true if n >= low and n <= high, considering the fact that 
// this is mod256 and wrapping must be considered
bool isWithinWindow(uint8_t n, uint8_t low, uint8_t high)
{
    if (high >= low)
    {
	// Not wrapped
	return (n >= low && n <= high);
    }
    else
    {
	// High has wrapped
	return (!(n < low && n > high));
    }
}

// IFs
bool Map27DataLayer::acknowledgement_inside_window(uint8_t n_r)
{
    DEBUG_PRINTF("acknowledgement_inside_window %d %d %d\n", n_r, _stored_acknowledged_rx_sequence_number, _send_state);
    return isWithinWindow(n_r, _stored_acknowledged_rx_sequence_number, _send_state);
}

bool Map27DataLayer::all_transmitted_packets_acknowledged(uint8_t n_r)
{
    DEBUG_PRINTF("all_transmitted_packets_acknowledged %d %d\n", n_r,_send_state);
    return n_r == _send_state;
}

bool Map27DataLayer::immediate_reply_requested(uint8_t ar)
{
    return (ar == 1);
}

bool Map27DataLayer::packet_out_of_sequence(uint8_t n_s)
{
    // A2.2.11
    // Returns 'true' if the sequence_number (N(S)) of the last received link transfer message is not
    // equal to the 'receive_state_variable' (V(R)).
    return (n_s != _receive_state);
}

bool Map27DataLayer::packet_outside_window(uint8_t n_s)
{
    // Returns ´true´ if the tx_sequence_number of the last received ´link transfer´ message is outside
    // the range of values given by the expression [(rx_sequence_number - stored receive_credit) to
    // (rx_sequence_number-1 + rx_credit_number)], these variables being the values of parameters in
    // the last transmitted ´link_acknowledge´messages.
    DEBUG_PRINTF("packet_outside_window: %d %d %d %d\n", n_s, _receive_state, _stored_receive_credit, _receive_credit);
    return !isWithinWindow(n_s, _receive_state - _stored_receive_credit, _receive_state - 1 + _receive_credit);
}

bool Map27DataLayer::receive_credit_available()
{
    // Returns 'true' if there is enough buffer space available in the network layer to pass the
    // network_layer_message content of the last received 'link_transfer' message to the network
    // layer.
    return (_receive_credit > 0);
}

bool Map27DataLayer::received_parameters_acceptable(uint8_t n1, uint8_t k, uint8_t version)
{
    return (   n1 <= MAP27_N1
	       && k <= MAP27_WINDOW_SIZE
	       && version <= MAP27_PROTOCOL_VERSION);
}

bool Map27DataLayer::repeated_link_acknowledge(uint8_t n_r)
{
    // Returns 'true' if the rx_sequence_number of the last received 'link_acknowledge' message is
    // equal to the rx_sequence_number of the last but one received 'link_acknowledge' message, and
    // is not equal to 'send_state'.
    return (   n_r == _second_last_n_r
	    && n_r != _send_state);
}

bool Map27DataLayer::retransmission_count_zero()
{
    return (_retransmission_count == 0);
}

bool Map27DataLayer::send_credit_available()
{
    DEBUG_PRINTF("send_credit_available %d\n", _send_credit);
    return (_send_credit > 0);
}

void Map27DataLayer::output_link_ready()
{
    DEBUG_PRINTF("output_link_ready\n");
    if (_networkLayer)
	_networkLayer->link_ready();
}

void Map27DataLayer::output_link_failure()
{
    DEBUG_PRINTF("output_link_failure\n");
    if (_networkLayer)
	_networkLayer->link_failure();
}

void Map27DataLayer::output_packet_accepted()
{
    DEBUG_PRINTF("output_packet_accepted\n");
    if (_networkLayer)
	_networkLayer->packet_accepted();
}

void Map27DataLayer::output_packet_rejected()
{
    DEBUG_PRINTF("output_packet_rejected\n");
    if (_networkLayer)
	_networkLayer->packet_rejected();
}

void Map27DataLayer::output_network_layer_packet(uint8_t* buf, uint16_t bufLen)
{
    DEBUG_PRINTF("output_network_layer_packet\n");
    if (_networkLayer)
	_networkLayer->network_layer_packet(buf, bufLen);
}

bool Map27DataLayer::isReady()
{
    return (_link_state == LINK_STATE_READY);
}

void Map27DataLayer::output_link_request()
{
    DEBUG_PRINTF("TX LR output_link_request %d %d %d\n", _maximum_length, _window_size, _version);
    // A2.2.6
    // This message is sent to the RS-232 port with its parameters set to the current values of
    // maximum_length, window_size and version.
    if (_physicalLayer)
	_physicalLayer->sendLinkRequest(_maximum_length, _window_size, _version);
}

void Map27DataLayer::output_link_acknowledge()
{
    DEBUG_PRINTF("TX LA output_link_acknowledge %d %d\n", _receive_state, _receive_credit);
    // A2.2.6
    // When this output is initiated, the data-link layer first stores the value of 'receive_credit' in
    // variable send_credit'. The link_acknowledge message is then sent to the RS-232 port with
    // parameter 'rx_sequence_number' given the value of 'receive_state' and parameter
    // 'rx_credit_number' given the value of 'receive_credit'.
    // CAUTION: NO: saving receive_credit in send_credit is WRONG

    if (_physicalLayer)
	_physicalLayer->sendLinkAcknowledge(_receive_state, _receive_credit);
}

void Map27DataLayer::output_link_transfer()
{
    DEBUG_PRINTF("TX LT output_link_transfer %d %d\n", _send_state,  _send_credit);
    // A2.2.6
    // When this message is sent to the RS-232 port, its parameters are set to 'send_state' and
    // 'LT_acknowledge_request' respectively, and its data area is filled with a network layer message
    // from the queue of messages waiting to be transmitted. The correct message is identified by
    // selecting the one whose stored_packet_number is equal to the value of the variable 'send_state'.
    DEBUG_DUMP("output_link_transfer", _txQueueBuf[_txNext], _txQueueBuflen[_txNext]);
    if (_physicalLayer)
	_physicalLayer->sendLinkTransfer(_send_state, _send_credit > 1 ? 0 : 1, _txQueueBuf[_txNext], _txQueueBuflen[_txNext]);
    _txNext = (_txNext + 1) % _window_size;
}

uint32_t Map27DataLayer::calc_retry_timeout()
{
    // Compute T1 in msec based on baud rate and N1 using table 4-1
    // X, Y position in table 4.1
    uint8_t x, y;

    // Times in milliseconds
    static uint32_t retry_timeout[6][4] = 
	{
	    //   0    10    100    more    // R
	    { 1300, 6700, 54700, 137300 }, // 600
	    { 700,  3400, 27400, 68700  }, // 1200
	    { 400,  1700, 13700, 24400  }, // 2400
	    { 200,  900,  6900,  17200  }, // 4800
	    { 100,  500,  3500,  8600   }, // 9600
	    { 50,   300,  1800,  4300   }  // more
	};

    if (_maximum_length == 0)
	x = 0;
    else if (_maximum_length <= 10)
	x = 1;
    else if (_maximum_length <= 100)
	x = 2;
    else
	x = 3;

    if (_baud <= 600)
	y = 0;
    else if (_baud <= 1200)
	y = 1;
    else if (_baud <= 2400)
	y = 2;
    else if (_baud <= 4800)
	y = 3;
    else if (_baud <= 9600)
	y = 4;
    else
	y = 5;

    return retry_timeout[y][x];
}

uint32_t Map27DataLayer::calc_acknowledgement_timeout()
{
    // Compute T2 in msec based on T1 and baud rate:
    // The maximum value of T2 is T1 minus twice the length of an LA packet (11 octets)
    return calc_retry_timeout() - (2 * 11 * 10 / _baud);
}

