// DataLayer.h
//
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2013 Mike McCauley
// $Id: DataLayer.h,v 1.3 2013/07/13 11:28:33 mikem Exp $

#ifndef MAP27_DATALAYER_H
#define MAP27_DATALAYER_H

#include <stdint.h>

#include "Map27.h"
#include "Timer.h"

class Map27PhysicalLayer;
class Map27NetworkLayer;

/////////////////////////////////////////////////////////////////////
/// \class Map27DataLayer DataLayer.h <Map27/DataLayer.h>
/// \brief Manage the Data Layer of the Map27 protocol stack
///
/// This class receives raw messages from Network LAyer for transmission to the radio.
/// It also receves low level frames from the Physical Layer and delivers them to Network Layer.
/// It is responsible for establishing and maintaining the protocol link between this protocol stack and the radio.
/// Based on Appendix A2
class Map27DataLayer
{
public:
    /// Constructor
    Map27DataLayer();

    /// Poll the Data Layer for timer and input events from lower layers in the protocol stack
    /// In Unix/Linux and Arduino environments, you are required to call this at least as 
    /// frequently as octets might be received from the radio. At least every 10msec would 
    /// be appropriate for a 9600 baud connection.
    /// Note: Map27NetworkLayer does this automatically when its poll() function is called.
    void poll();
    
    /// Sets the pointer to the PhysicalLayer to use
    /// to communicate with the Radio
    /// \param[in] physicalLayer Pointer to the Physical Layer
    void setPhysicalLayer(Map27PhysicalLayer* physicalLayer);

    /// Get the current Physical Layer pointer
    /// \return The current Physical Layer pointer, or NULL if previously set.
    Map27PhysicalLayer* physicalLayer();
    
    /// Sets the pointer to the NetworkLayer to use
    /// to deliver messages up the protocol stack
    /// \param[in] networkLayer Pointer to the Netowrk Layer
    void setNetworkLayer(Map27NetworkLayer* networkLayer);

    /// Get the current Network Layer pointer
    /// \return The current Network Layer pointer, or NULL if previously set.
    Map27NetworkLayer* networkLayer();
    
    /// Sets the desired baud rate to communicate
    /// with the radio. Typical choices are 9600 or 1200, but other speeds may be suported by your radio 
    /// and RS232 hardware
    /// \param[in] baud The desired baud rate in baud (per second)

    void setBaud(uint32_t baud);

    /// Get the current 
    /// baud rate
    /// \return The current baud rate in baud (per second)
    uint32_t baud();
    
    /// \name Inputs from OS A2.2.2

    /// Powers on the Data LAyer and lower layers.
    /// Opens the port aand initialises it.
    /// initialises the Physical layer
    void power_on();

    /// \name Inputs from PhysicalLayer A2.2.4
    /// These members are called by the Physical LAyer in response toe events detected by it

    /// Called by Physical Layer
    /// when a Linkrequest (LR) message is received by the radio
    /// \param[in] n1 Maximum message length
    /// \param[in] k Window size
    /// \param[in] version Protocol version
    void link_request(uint8_t n1, uint8_t k, uint8_t version);

    /// Called by Physical Layer
    /// when a Link Acknowledge (LA) message is received by the radio
    /// \param[in] n_r Receive sequence number
    /// \param[in] n_k receivet credit number
    void link_acknowledge(uint8_t n_r, uint8_t n_k);

    /// Called by Physical Layer
    /// when a Link Transfer (LT) message is received by the radio
    /// \param[in] n_s Transmit sequence number
    /// \param[in] ar Immediate Acknowledgement request
    /// \param[in] buf Pointer to the payload data
    /// \param[in] bufLen Number of bytes in buf
    void link_transfer(uint8_t n_s, uint8_t ar, uint8_t* buf, uint16_t bufLen);

    /// \name Inputs from NetworkLayer
    /// These members are called by the Network layer to control
    /// the Data Layer and queue mesage for transmission to the radio

    /// Called by Network Layer wen it requires to 
    /// reset the Data Layer
    void network_layer_reset();

    /// Called by the Network Layer
    /// whenever it completes handling of a message and is able to handle more
    /// \param[in] r_k Number of concurrent messages the Network Layer can now handle
    void credit_value(uint8_t r_k);

    /// Called by the Network Layer
    /// to queue a raw message for delivery to the radio
    /// \param[in] buf Pointer to the message
    /// \param[in] bufLen Number of bytes in buf
    /// \return true if the link is READY and the message was queued for delivery
    bool network_layer_packet(uint8_t* buf, uint16_t bufLen);

    /// Tells whether the Data Layer is READY
    /// and therefor whether the link to teh radio is established
    /// \return tru if the Daa Layer is in READY state
    bool isReady();

protected:
    void reset();
    void resetLink();

    // Timer events A2.2.3
    void link_establishment_timeout();     // T0
    void retry_timeout();                  // T1
    void acknowledgement_timeout();        // T2
    void activity_timeout();               // T3
    void link_failure_detection_timeout(); // T4

    // Tasks A2.2.10
    void adjust_link_parameters(uint8_t n1, uint8_t k, uint8_t version);
    void decrement_retransmission_counter();
    void decrement_send_credit();
    void delete_acknowledged_packets(uint8_t n_r);
    void increment_send_state();
    void initialise_variables();
    void initialise_rs232_port();
    void maximise_link_parameters();
    void record_send_credit(uint8_t n_r, uint8_t n_k);
    void rewind_packet_number(uint8_t n_r);
    void set_retransmission_counter(uint8_t c1xo);
    void store_acknowledged_rx_sequence_number(uint8_t n_r);
    void store_packet(uint8_t* buf, uint16_t bufLen);
    void store_receive_credit();
    void update_receive_credit(uint8_t r_k);
    // Missing from list:
    void increment_receive_state();
    void decrement_receive_credit();

    // IFs A2.2.11
    bool acknowledgement_inside_window(uint8_t n_r);
    bool all_transmitted_packets_acknowledged(uint8_t n_r);
    bool immediate_reply_requested(uint8_t ar);
    bool packet_out_of_sequence(uint8_t n_s);
    bool packet_outside_window(uint8_t n);
    bool receive_credit_available();
    bool received_parameters_acceptable(uint8_t n1, uint8_t k, uint8_t version);
    bool repeated_link_acknowledge(uint8_t n_r);
    bool retransmission_count_zero();
    bool send_credit_available();

    /// Outputs to Network Layer A2.2.7
    virtual void output_link_ready();
    virtual void output_link_failure();
    virtual void output_packet_accepted();
    virtual void output_packet_rejected();
    virtual void output_network_layer_packet(uint8_t* buf, uint16_t bufLen);

    // Outputs to Physical Layer A2.2.6
    virtual void output_link_request();
    virtual void output_link_acknowledge();
    virtual void output_link_transfer();

    // Callback glue
    static void link_establishment_timer_callback_glue(void* target);
    static void retry_timer_callback_glue(void* target);
    static void acknowledgement_timer_callback_glue(void* target);
    static void activity_timer_callback_glue(void* target);
    static void link_failure_detection_timer_callback_glue(void* target);

    uint32_t calc_retry_timeout();
    uint32_t calc_acknowledgement_timeout();

private:
    typedef enum
    {
	LINK_STATE_RESET_WAIT = 1,
	LINK_STATE_LINK_WAIT,
	LINK_STATE_READY,
    } LinkState;

    Map27PhysicalLayer* _physicalLayer;
    Map27NetworkLayer*  _networkLayer;

    uint32_t            _baud;
    LinkState           _link_state;     // Current state of Data Layer

    // Negotiated link parameter variables
    uint8_t             _maximum_length; // Negotiated N1
    uint8_t             _window_size;    // Neotiated k
    uint8_t             _version;        // Negotiated VERSION

    // Variables and sequence numbers 4.3.2.4
    uint8_t             _send_state;     // V(S)
    uint8_t             _receive_state;  // V(R)
    uint8_t             _receive_credit; // R(k)
    uint8_t             _send_credit;    // S(k)
    uint8_t             _stored_acknowledged_rx_sequence_number; // SN(R)
    uint8_t             _stored_receive_credit; // SR(k);
    uint8_t             _retransmission_count; // C1

    uint8_t             _last_n_r; // rx_sequence_number (n_r) from last LA
    uint8_t             _second_last_n_r; // rx_sequence_number (n_r) from second last LA

    // Timers
    Map27Timer          _link_establishment_timer;
    Map27Timer          _retry_timer;
    Map27Timer          _acknowledgement_timer;
    Map27Timer          _activity_timer;
    Map27Timer          _link_failure_detection_timer;

    // Transmitter message queue
    uint8_t             _txQueueBuf[MAP27_MAX_MESSAGE_LEN][MAP27_WINDOW_SIZE]; // Message contents
    uint8_t             _txQueueBuflen[MAP27_WINDOW_SIZE];
    uint8_t             _txQueueLength; // Number of messages in queue
    uint8_t             _txBottom; // Index of oldest unacknowledged message in queue
    uint8_t             _txNext;   // Index of next message to be sent.
};

#endif
