// Map27.h
//
// Basic definitions for Map27 library
//
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2013 Mike McCauley
// $Id: DataLayer.h,v 1.1 2013/07/09 23:38:09 mikem Exp mikem $

/// \mainpage Map27 library
///
/// This is the Map27 library.
/// It provides a C++ object-oriented interface for sending and receiving message to 
/// a Map27 enabled trunked radio.
///
/// Map27 (http://en.wikipedia.org/wiki/Mobile_Access_Protocol 
/// specifies an interface between a trunked mobile radio and a data terminal equipment.
/// This library is based on "MAP27 Mobile Access Protocol for MPT 1327 equipment Version 1.5" 
/// (http://www.mpt1327-usa.org/downloads/map27.pdf)
/// from the User Access Definition Group.
///
/// This library, along with a Map27 comnpatible trunlked radio can be used to set up voice and
/// data calls between radios, and to monoitor the activity of the radio.
///
/// The specification defines the following ISO conforming network layers:
/// - Physical Layer
/// - Data Layer
/// - Network Layer
/// - Application Layer
/// and this Map27 library provides classes to implement each layer, and to interconnect them to form a complete
/// application and protocol stack.
///
/// Map27 library has been tested on Unix (OpenSuSE 12.3) and should run on any modern linux. 
/// It has been tested against:
/// - Tait 2040 trunked radio
///
/// It has been designed to be readily portable to other architectures and OSs, including QT and Arduino 
/// (athough this has not yet been done)
///
/// The latest version of this documentation can be downloaded from 
/// http://www.airspayce.com/mikem/Map27
/// The version of the package that this documentation refers to can be downloaded 
/// from http://www.airspayce.com/mikem/Map27/Map27-1.3.tgz
///
/// \par Overall Architecture
///
/// The usual way to construct a complete ap27 application is to create your own subclass of 
/// Map27ApplicationLayer, then interconnect it to the other protocol layer classes like this:
/// \code
/// [Your subclass of Map27ApplicationLayer]
///        ^   |
///        |   v
/// [ Map27NetworkLayer ]
///        ^   |
///        |   v
/// [ Map27DataLayer ]
///        ^   |
///        |   v
/// [ Map27PhysicalLayer ]
///        ^   |
///        |   v
/// [ subclass of Map27Port, such as Map27UnixPort ]
///        ^   |
///        |   v
/// [ RS232 connection ]
///        ^   |
///        |   v
/// [ Map27 enabled trunked radio ]
/// \endcode
/// 
/// \par Logging
///
/// A simple logging facility is provided in Log.h and Log.cpp, 
/// which can be customised and overridden to suit your needs.
///
/// \par Licensing
///
/// This software is Copyright (C) 2013 Mike McCauley. Use is subject to license
/// conditions. The main licensing options available are GPL V2 or Commercial:
/// 
/// \par Open Source Licensing GPL V2
///
/// This is the appropriate option if you want to share the source code of your
/// application with everyone you distribute it to, and you also want to give them
/// the right to share who uses it. If you wish to use this software under Open
/// Source Licensing, you must contribute all your source code to the open source
/// community in accordance with the GPL Version 2 when your application is
/// distributed. See http://www.gnu.org/copyleft/gpl.html
/// 
/// \par Commercial Licensing
///
/// This is the appropriate option if you are creating proprietary applications
/// and you are not prepared to distribute and share the source code of your
/// application. Contact info@airspayce.com for details.
///
/// \par Revision History
///
/// \version 1.1 Initial release
///
/// \version 1.2 test3.ccpp was missing from 1.1 distribution
//
/// \version 1.3 2011-12-8 Fixed incorrect distro link.<br>
///
/// @example test.cpp
/// Test suite that exercises and tests much of the code without being connected to a real radio
///
/// @example test2.cpp
/// Test program used to exercise the code when connected to a real radio
///
/// @example test3.cpp
/// Sample program showing how to create a real application that connects to a real radio.
/// This sample connects to /dev/ttyUSB0 at 9600 baud and sends a radio personality
/// request every second. It prints out the details of the reply.

#ifndef MAP27_MAP27_H
#define MAP27_MAP27_H

// The version of the Map27 protocol we can handle
// Corresponds to "MAP27 Mobile Access Protocol for MPT 1327 equipment Version 1.5" January 1998
#define MAP27_PROTOCOL_VERSION 1

// Default Baud rate
#define MAP27_DEFAULT_BAUD 9600

// N1 4.5.6
// Max message legth this uint can handle
// Actual length is negotiated at link establishment time
#define MAP27_N1                   10
#define MAP27_MAX_MESSAGE_LEN      (16*(MAP27_N1+1))

// k 4.3.2.1
// Max number of pending messages this unit can handle
// MAx window size k
#define MAP27_WINDOW_SIZE      5

// N2 4.5.7
#define MAP27_MAX_RETRANSMISSIONS 10

// N3 4.5.8
#define MAP27_MAX_ACTIVITY_TIMEOUTS 5

// T0 4.5.1
#define MAP27_ESTABLISHMENT_TIMEOUT 100
// T1 is computed by calc_retry_timeout()
// T2 is computed by calc_acknowledgement_timeout()
// T3 4.5.4:
#define MAP27_ACTIVITY_TIMEOUT      15000
// T4 4.5.5
#define MAP27_LINK_FAILURE_DETECTION_TIMEOUT ((MAP27_MAX_ACTIVITY_TIMEOUTS * MAP27_ACTIVITY_TIMEOUT) + 10)

#define min(a,b) ((a)<(b)?(a):(b))
#define max(a,b) ((a)>(b)?(a):(b))


#endif
