// NetworkLayer.cpp
//
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2013 Mike McCauley
// $Id: NetworkLayer.cpp,v 1.7 2013/07/14 06:42:42 mikem Exp $

// for memcpy():
#include <string.h>

#include "Address.h"
#include "NetworkLayer.h"
#include "DataLayer.h"
#include "ApplicationLayer.h"
#include "Log.h"

// Timer callback glue
void Map27NetworkLayer::request_timer_callback_glue(void* target)
{
    ((Map27NetworkLayer*)target)->request_timeout();
}

Map27NetworkLayer::Map27NetworkLayer()
{
    _dataLayer = NULL;
    _applicationLayer = NULL;
    // Set up timer callbacks
    _request_timer.setCallback(request_timer_callback_glue, this);
    reset();
}

void Map27NetworkLayer::setDataLayer(Map27DataLayer* dataLayer)
{
    _dataLayer = dataLayer;
}

Map27DataLayer* Map27NetworkLayer::dataLayer()
{
    return _dataLayer;
}

void Map27NetworkLayer::setApplicationLayer(Map27ApplicationLayer* applicationLayer)
{
    _applicationLayer = applicationLayer;
}

Map27ApplicationLayer* Map27NetworkLayer::applicationLayer()
{
    return _applicationLayer;
}

Map27NetworkLayer::NetworkState Map27NetworkLayer::state()
{
    return _network_state;
}

bool Map27NetworkLayer::is_idle()
{
    return _network_state == NETWORK_STATE_IDLE;
}

void Map27NetworkLayer::power_on()
{
    if (_dataLayer)
	_dataLayer->power_on();
}

void Map27NetworkLayer::reset()
{
    _network_state = NETWORK_STATE_IDLE;
    _request_timer.stop();
}

void Map27NetworkLayer::poll()
{
    if (_dataLayer)
	_dataLayer->poll();
}

// Inputs from Data Layer:
void Map27NetworkLayer::link_ready()
{
    DEBUG_PRINTF("NetworkLayer link_ready\n");
    if (_applicationLayer)
	_applicationLayer->link_ready();
}

void Map27NetworkLayer::link_failure()
{
    DEBUG_PRINTF("NetworkLayer link_failure\n");
    reset();
    if (_applicationLayer)
	_applicationLayer->link_failure();
}

void Map27NetworkLayer::packet_accepted()
{
    DEBUG_PRINTF("NetworkLayer packet_accepted\n");
    if (_applicationLayer)
	_applicationLayer->packet_accepted();
}

void Map27NetworkLayer::packet_rejected()
{
    DEBUG_PRINTF("NetworkLayer packet_rejected\n");
    if (_applicationLayer)
	_applicationLayer->packet_rejected();
}

void Map27NetworkLayer::network_layer_packet(uint8_t* buf, uint16_t bufLen)
{
    DEBUG_DUMP("NetworkLayer: received packet", buf, bufLen);

    decode_network_layer_packet(buf, bufLen);

    // Message received by DataLayer, collect it and remove it,
    // making room in the data layer for another
    if (_dataLayer)
	_dataLayer->credit_value(1);
}

// Timer Events
void Map27NetworkLayer::request_timeout()
{
    DEBUG_PRINTF("request_timeout %d. IDLE\n", _network_state);
    _network_state = NETWORK_STATE_IDLE;
    output_application_error(MAP27_APPLICATION_ERROR_REQUEST_TIMEOUT);
}

void Map27NetworkLayer::decode_network_layer_packet(uint8_t* buf, uint16_t bufLen)
{
    // Handle received network messages
    // First octet is message type
    if (bufLen)
    {
	switch (buf[0])
	{
	    case MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_STATUS:
		receiveStatus(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_ACK_POSITIVE:
	    case MAP27_NETWORK_MESSAGE_TYPE_ACK_QUEUEING:
	    case MAP27_NETWORK_MESSAGE_TYPE_ACK_NEGATIVE:
		// Also handles MST Ack
		receiveSSTAck(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_SST:
		receiveSST(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_MST:
		receiveMST(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_SETUP_PROGRESS_POSITIVE:
	    case MAP27_NETWORK_MESSAGE_TYPE_SETUP_PROGRESS_QUEUEING:
	    case MAP27_NETWORK_MESSAGE_TYPE_SETUP_PROGRESS_NEGATIVE:
		receiveSetupProgress(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_INCOMING_VOICE_MODEM:
	    case MAP27_NETWORK_MESSAGE_TYPE_INCOMING_EMERGENCY_VOICE_MODEM:
		receiveIncomingVoiceModem(buf, bufLen); // Handles emergency too
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_PROGRESS_POSITIVE:
	    case MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_PROGRESS_WARNING:
	    case MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_PROGRESS_NEGATIVE:
		receiveProgress(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_MODEM_DATA:
		receiveModemData(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_CLEARED_NORMAL:
	    case MAP27_NETWORK_MESSAGE_TYPE_CLEARED_ABNORMAL:
		// Tait 2040 sends this (a6 07 17 6c 04) when PTT is pressed and not home
		receiveCleared(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_DIVERSION_ACK_POSITIVE:
	    case MAP27_NETWORK_MESSAGE_TYPE_DIVERSION_ACK_NEGATIVE:
	    case MAP27_NETWORK_MESSAGE_TYPE_RADIO_PERSONALITY:
		receiveRadioPersonality(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_NUMBERING_INFORMATION:
		// Tait 2040 sends in repsonse to Radio Interrogation (Numbering):
		// (b1 07 17 6c 95 e2 18 9d 98 9e 19 00 07 18 a4 07 18 a3 07 18 a2 07 18 a1 07 18 a0 07 18 9f 07 18 9e)
		receiveNumberingInfo(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_OPERATING_CONDITION:
		// Tait 2040 sends this (b2 03 00 ff), (b2 02 00 ff) 
		// when PPT pressed and conventional channel selected
		receiveOperatingCondition(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_NETWORK_INFORMATION:
		// Tait 2040 sends: (b5 00 00 00 00)
		receiveNetworkInformation(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_RADIO_SETTINGS:
		// Caution, Tait 204 sends short version without codings: (b3 f4 00 00)
		receiveRadioSettings(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_PROTOCOL_INFO:
		// Tait 2040 sends this (b4 01) when try to send status when not home
		receiveProtocolInfo(buf, bufLen);
		break;

	    case MAP27_NETWORK_MESSAGE_TYPE_RADIO_TEST_RESULT:
		receiveRadioTestResult(buf, bufLen);
		break;

	    default:
		DEBUG_PRINTF("NetworkLayer: Unknown message type: %02x\n", buf[0]);
		break;

		
	}
    }
}

///////////////////////////////////////////////////////
// Send messages
bool Map27NetworkLayer::sendStatus(const Map27Address& address, uint8_t status)
{
    if (_network_state == NETWORK_STATE_IDLE)
    {
	uint8_t msg[MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN];
	msg[0] = MAP27_NETWORK_MESSAGE_TYPE_SEND_STATUS;
	uint8_t alen = address.encode(msg+1, sizeof(msg) - 1);
	msg[alen+1] = status & 0x1f; // Only 0..31 permitted
	
	if (_dataLayer)
	{
	    if (_dataLayer->network_layer_packet(msg, alen + 2))
	    {
		_network_state = NETWORK_STATE_STATUS_WAIT;
		_request_timer.start(MAP27_NETWORK_REQUEST_TIMEOUT);
		return true;
	    }
	}
	return false;
    }
    else
    {
	DEBUG_PRINTF("NetworkLayer: sendStatus: invalid network state %d\n", _network_state);
	output_application_error(MAP27_APPLICATION_ERROR_INVALID_STATE);
	return false;
    }
}

bool Map27NetworkLayer::sendRadioInterrogation(uint8_t reason)
{
    MsgRadioInterrogation msg;
    msg.type = MAP27_NETWORK_MESSAGE_TYPE_RADIO_INTERROGATION;
    msg.reason = reason;
    if (_dataLayer)
    {
	if (_dataLayer->network_layer_packet((uint8_t*)&msg, sizeof(msg)))
	{
	    // No change to state
	    return true;
	}
    }
    return false;
}

bool Map27NetworkLayer::sendSST(const Map27Address& address, uint8_t* buf, uint16_t bufLen)
{
    if (_network_state == NETWORK_STATE_IDLE)
    {
	uint8_t msg[MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN];

	msg[0] = MAP27_NETWORK_MESSAGE_TYPE_SEND_SST;
	uint8_t alen = address.encode(msg + 1, sizeof(msg) - 1);
	if (alen + 1 + bufLen > MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN)
	    return false;
	memcpy(msg + alen + 1, buf, bufLen); // Includes coding/NLB in first octet
	if (_dataLayer)
	{
	    if (_dataLayer->network_layer_packet(msg, bufLen + alen + 1))
	    {
		_network_state = NETWORK_STATE_SST_WAIT;
		_request_timer.start(MAP27_NETWORK_REQUEST_TIMEOUT);
		return true;
	    }
	}
	return false;
    }
    else
    {
	DEBUG_PRINTF("NetworkLayer: sendSST: invalid network state %d\n", _network_state);
	output_application_error(MAP27_APPLICATION_ERROR_INVALID_STATE);
	return false;
    }
}

bool Map27NetworkLayer::sendMST(const Map27Address& address, uint8_t* buf, uint16_t bufLen)
{
    if (_network_state == NETWORK_STATE_IDLE)
    {
	uint8_t msg[MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN];

	msg[0] = MAP27_NETWORK_MESSAGE_TYPE_SEND_MST;
	uint8_t alen = address.encode(msg + 1, sizeof(msg) - 1);
	if (alen + 1 + bufLen > MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN)
	    return false;
	memcpy(msg + alen + 1, buf, bufLen); // Includes coding/NLB in first octet
	
	if (_dataLayer)
	{
	    if (_dataLayer->network_layer_packet(msg, bufLen + alen + 1))
	    {
		_network_state = NETWORK_STATE_MST_WAIT;
		_request_timer.start(MAP27_NETWORK_REQUEST_TIMEOUT);
		return true;
	    }
	}
	return false;
    }
    else
    {
	DEBUG_PRINTF("NetworkLayer: sendMST: invalid network state %d\n", _network_state);
	output_application_error(MAP27_APPLICATION_ERROR_INVALID_STATE);
	return false;
    }
}

bool Map27NetworkLayer::setupVoiceModem(const Map27Address& address, uint8_t details)
{
    if (   _network_state == NETWORK_STATE_IDLE
	|| _network_state == NETWORK_STATE_ACTIVE)
    {
	uint8_t msg[MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN];

	msg[0] = MAP27_NETWORK_MESSAGE_TYPE_SETUP_VOICE_MODEM;
	uint8_t alen = address.encode(msg + 1, sizeof(msg) - 1);
	msg[alen+1] = details;
	if (_dataLayer)
	{
	    if (_dataLayer->network_layer_packet(msg, alen + 2))
	    {
		if (_network_state == NETWORK_STATE_IDLE)
		    _network_state = NETWORK_STATE_OUTCALL_WAIT;
		else
		    _network_state = NETWORK_STATE_OUTCALL_INCLUDE_WAIT;
		_request_timer.start(MAP27_NETWORK_REQUEST_TIMEOUT);
		return true;
	    }
	}
	return false;
    }
    else
    {
	DEBUG_PRINTF("NetworkLayer: setupVoiceModem: invalid network state %d\n", _network_state);
	output_application_error(MAP27_APPLICATION_ERROR_INVALID_STATE);
	return false;
    }
}

bool Map27NetworkLayer::setupEmergencyVoiceModem(const Map27Address& address, uint8_t details)
{
    if (   _network_state == NETWORK_STATE_IDLE
	|| _network_state == NETWORK_STATE_ACTIVE)
    {
	uint8_t msg[MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN];

	msg[0] = MAP27_NETWORK_MESSAGE_TYPE_SETUP_EMERGENCY_VOICE_MODEM;
	uint8_t alen = address.encode(msg + 1, sizeof(msg) - 1);
	msg[alen+1] = details;
	if (_dataLayer)
	{
	    if (_dataLayer->network_layer_packet(msg, alen + 2))
	    {
		if (_network_state == NETWORK_STATE_IDLE)
		    _network_state = NETWORK_STATE_OUTCALL_WAIT;
		else
		    _network_state = NETWORK_STATE_OUTCALL_INCLUDE_WAIT;
		_request_timer.start(MAP27_NETWORK_REQUEST_TIMEOUT);
		return true;
	    }
	}
	return false;
    }
    else
    {
	DEBUG_PRINTF("NetworkLayer: setupEmergencyVoiceModem: invalid network state %d\n", _network_state);
	output_application_error(MAP27_APPLICATION_ERROR_INVALID_STATE);
	return false;
    }
}

bool Map27NetworkLayer::sendModemData(uint8_t* buf, uint16_t bufLen)
{
    uint8_t msg[MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN];
    if (bufLen <= sizeof(msg) - 1)
    {
	msg[0] = MAP27_NETWORK_MESSAGE_TYPE_SEND_MODEM_DATA;
    	memcpy(msg + 1, buf, bufLen);
	
	if (_dataLayer)
	{
	    if (_dataLayer->network_layer_packet(msg, bufLen + 1))
	    {
		_network_state = NETWORK_STATE_MST_WAIT;
		_request_timer.start(MAP27_NETWORK_REQUEST_TIMEOUT);
		return true;
	    }
	}
    }
    return false;
}

bool Map27NetworkLayer::sendDisconnect(const Map27Address& address, uint8_t cause)
{
    uint8_t msg[MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN];
    msg[0] = cause == MAP27_NETWORK_MESSAGE_DISCONNECT_CAUSE_NORMAL
	? MAP27_NETWORK_MESSAGE_TYPE_DISCONNECT_NORMAL : MAP27_NETWORK_MESSAGE_TYPE_DISCONNECT_CANCEL;
    uint8_t alen = address.encode(msg+1, sizeof(msg) - 1);
    msg[alen+1] = cause;
    
    if (_dataLayer)
    {
	if (_dataLayer->network_layer_packet(msg, alen + 2))
	{
	    _network_state = NETWORK_STATE_IDLE;
	    _request_timer.stop();
	    return true;
	}
    }
    return false;
}

bool Map27NetworkLayer::sendCleared(const Map27Address& address, uint8_t cause)
{
    uint8_t msg[MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN];
    msg[0] = cause == MAP27_NETWORK_MESSAGE_CLEARED_CAUSE_NORMAL
	? MAP27_NETWORK_MESSAGE_TYPE_CLEARED_NORMAL : MAP27_NETWORK_MESSAGE_TYPE_CLEARED_ABNORMAL;
    uint8_t alen = address.encode(msg+1, sizeof(msg) - 1);
    msg[alen+1] = cause;
    
    if (_dataLayer)
    {
	if (_dataLayer->network_layer_packet(msg, alen + 2))
	{
	    _network_state = NETWORK_STATE_IDLE;
	    _request_timer.stop();
	    return true;
	}
    }
    return false;
}

bool Map27NetworkLayer::sendDiversionRequest(const Map27Address& address, uint8_t info, const Map27Address& diversionAddress)
{
    if (_network_state == NETWORK_STATE_IDLE)
    {
	uint8_t msg[MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN];

	msg[0] = MAP27_NETWORK_MESSAGE_TYPE_DIVERSION_REQUEST;
	uint8_t alen = address.encode(msg + 1, sizeof(msg) - 1);
	msg[alen+1] = info;
	uint8_t alen2 = diversionAddress.encode(msg + alen + 2, sizeof(msg) - alen + 2);

	if (_dataLayer)
	{
	    if (_dataLayer->network_layer_packet(msg, alen + 2 + alen2))
	    {
		_network_state = NETWORK_STATE_DIVERSION_WAIT;
		_request_timer.start(MAP27_NETWORK_REQUEST_TIMEOUT);
		return true;
	    }
	}
	return false;
    }
    else
    {
	DEBUG_PRINTF("NetworkLayer: sendDiversionRequest: invalid network state %d\n", _network_state);
	output_application_error(MAP27_APPLICATION_ERROR_INVALID_STATE);
	return false;
    }
}

bool Map27NetworkLayer::sendDiversionCancel(const Map27Address& address, uint8_t info)
{
    if (_network_state == NETWORK_STATE_IDLE)
    {
	uint8_t msg[MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN];

	msg[0] = MAP27_NETWORK_MESSAGE_TYPE_DIVERSION_CANCEL;
	uint8_t alen = address.encode(msg + 1, sizeof(msg) - 1);
	msg[alen+1] = info;

	if (_dataLayer)
	{
	    if (_dataLayer->network_layer_packet(msg, alen + 2))
	    {
		_network_state = NETWORK_STATE_DIVERSION_WAIT;
		_request_timer.start(MAP27_NETWORK_REQUEST_TIMEOUT);
		return true;
	    }
	}
	return false;
    }
    else
    {
	DEBUG_PRINTF("NetworkLayer: sendDiversionRequest: invalid network state %d\n", _network_state);
	output_application_error(MAP27_APPLICATION_ERROR_INVALID_STATE);
	return false;
    }
}

bool Map27NetworkLayer::sendVolumeControl(uint8_t control_port, uint8_t volume)
{
    uint8_t msg[MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN];
    msg[0] = MAP27_NETWORK_MESSAGE_TYPE_VOLUME_CONTROL;
    msg[1] = control_port;
    msg[2] = volume;

    if (_dataLayer)
	if (_dataLayer->network_layer_packet(msg, 3))
	    return true;
    return false;
}

bool Map27NetworkLayer::sendDialledString(uint8_t* buf, uint16_t bufLen)
{
    uint8_t msg[MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN];
    if (bufLen <= sizeof(msg) - 1)
    {
	msg[0] = MAP27_NETWORK_MESSAGE_TYPE_DIALLED_STRING;
    	memcpy(msg + 1, buf, bufLen);
	
	if (_dataLayer)
	{
	    if (_dataLayer->network_layer_packet(msg, bufLen + 1))
	    {
		return true;
	    }
	}
    }
    return false;
}

bool Map27NetworkLayer::sendRadioTest(uint8_t* buf, uint16_t bufLen)
{
    uint8_t msg[MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN];
    if (bufLen <= sizeof(msg) - 1)
    {
	msg[0] = MAP27_NETWORK_MESSAGE_TYPE_RADIO_TEST;
    	memcpy(msg + 1, buf, bufLen);
	
	if (_dataLayer)
	{
	    if (_dataLayer->network_layer_packet(msg, bufLen + 1))
	    {
		return true;
	    }
	}
    }
    return false;
}

///////////////////////////////////////////////////////
// Handle received messages
void Map27NetworkLayer::receiveStatus(uint8_t* buf, uint16_t bufLen)
{
    // 5.2.2.1.2
    if (bufLen >= 6)
    {
	Map27Address address;
	uint8_t alen = address.decode(buf + 1, bufLen - 1);
	if (bufLen >= alen + 2)
	{
	    uint8_t status = buf[alen+1];
	    DEBUG_PRINTF("NetworkLayer: receiveStatus %d. IDLE\n", status);
	    _network_state = NETWORK_STATE_IDLE;
	    _request_timer.stop();
	    if (_applicationLayer)
		_applicationLayer->receiveStatus(address, status);
	}
    }

}

void Map27NetworkLayer::receiveStatusAck(uint8_t* buf, uint16_t bufLen)
{
    // 5.2.2.1.3
    if (bufLen >= 6)
    {
	Map27Address address;
	uint8_t alen = address.decode(buf + 1, bufLen - 1);
	if (bufLen >= alen + 2)
	{
	    uint8_t cause = buf[alen+1];
	    DEBUG_PRINTF("NetworkLayer: receiveStatusAck %d. IDLE\n", cause);
	    _network_state = NETWORK_STATE_IDLE;
	    _request_timer.stop();
	    if (_applicationLayer)
		_applicationLayer->receiveStatusAck(address, cause);
	}
    }
}

void Map27NetworkLayer::receiveCleared(uint8_t* buf, uint16_t bufLen)
{
    // 5.2.2.6.2
    if (bufLen == 5)
    {
	Map27Address address;
	uint8_t alen = address.decodeBasic(buf + 1, bufLen - 1);
	if (bufLen >= alen + 2)
	{
	    uint8_t cause = buf[alen+1];
	    DEBUG_PRINTF("NetworkLayer: receiveCleared %d. IDLE\n", cause);
	    _network_state = NETWORK_STATE_IDLE;
	    _request_timer.stop();
	    if (_applicationLayer)
		_applicationLayer->receiveCleared(address, cause);
	}
    }
}

void Map27NetworkLayer::receiveSST(uint8_t* buf, uint16_t bufLen)
{
    // 5.2.2.2.2
    if (bufLen >= 6)
    {
	Map27Address address;
	uint8_t alen = address.decode(buf + 1, bufLen - 1);
	if (bufLen >= alen + 2)
	{
	    uint8_t cause = buf[alen+1];
	    DEBUG_PRINTF("NetworkLayer: receiveSST. IDLE\n", cause);
	    // No change to state
	    if (_applicationLayer)
		_applicationLayer->receiveSST(address, buf + alen + 1, bufLen - alen - 1);
	}
    }
}

void Map27NetworkLayer::receiveSSTAck(uint8_t* buf, uint16_t bufLen)
{
    // 5.2.2.2.3
    if (bufLen >= 6)
    {
	Map27Address address;
	uint8_t alen = address.decode(buf + 1, bufLen - 1);
	if (bufLen >= alen + 1)
	{
	    uint8_t cause = buf[alen+1];
	    DEBUG_PRINTF("NetworkLayer: receiveSSTAck %d. IDLE\n", cause);
	    _network_state = NETWORK_STATE_IDLE;
	    _request_timer.stop();
	    if (_applicationLayer)
		_applicationLayer->receiveSSTAck(address, cause);
	}
    }
}

void Map27NetworkLayer::receiveMST(uint8_t* buf, uint16_t bufLen)
{
    // 5.2.2.3.2
    if (bufLen >= 6)
    {
	Map27Address address;
	uint8_t alen = address.decode(buf + 1, bufLen - 1);
	if (bufLen >= alen + 2)
	{
	    uint8_t cause = buf[alen+1];
	    DEBUG_PRINTF("NetworkLayer: receiveMST. IDLE\n", cause);
	    // No change to state
	    if (_applicationLayer)
		_applicationLayer->receiveMST(address, buf + alen + 1, bufLen - alen - 1);
	}
    }
}

void Map27NetworkLayer::receiveMSTAck(uint8_t* buf, uint16_t bufLen)
{
    // 5.2.2.3.3
    if (bufLen >= 6)
    {
	Map27Address address;
	uint8_t alen = address.decode(buf + 1, bufLen - 1);
	if (bufLen >= alen + 2)
	{
	    uint8_t cause = buf[alen+1];
	    DEBUG_PRINTF("NetworkLayer: receiveMSTAck %d. IDLE\n", cause);
	    _network_state = NETWORK_STATE_IDLE;
	    _request_timer.stop();
	    if (_applicationLayer)
		_applicationLayer->receiveMSTAck(address, cause);
	}
    }
}

void Map27NetworkLayer::receiveSetupProgress(uint8_t* buf, uint16_t bufLen)
{
    // 5.2.2.4.3
    if (bufLen >= 6)
    {
	uint8_t type = buf[0];
	Map27Address address;
	uint8_t alen = address.decode(buf + 1, bufLen - 1);
	if (bufLen >= alen + 2)
	{
	    uint8_t cause = buf[alen+1];
	    DEBUG_PRINTF("NetworkLayer: receiveSetupProgress %d\n", cause);
	    if (_network_state == NETWORK_STATE_OUTCALL_WAIT)
	    {
		if (type == MAP27_NETWORK_MESSAGE_TYPE_SETUP_PROGRESS_POSITIVE)
		    _network_state == NETWORK_STATE_ACTIVE;
		if (type == MAP27_NETWORK_MESSAGE_TYPE_SETUP_PROGRESS_NEGATIVE)
		    _network_state == NETWORK_STATE_IDLE;
	    }
	    if (_applicationLayer)
		_applicationLayer->receiveSetupProgress(address, cause);
	}
    }
}

void Map27NetworkLayer::receiveIncomingVoiceModem(uint8_t* buf, uint16_t bufLen)
{
    // 5.2.2.4.4
    if (bufLen >= 6)
    {
	uint8_t type = buf[0];
	Map27Address address;
	uint8_t alen = address.decode(buf + 1, bufLen - 1);
	if (bufLen >= alen + 2)
	{
	    uint8_t details = buf[alen+1];
	    DEBUG_PRINTF("NetworkLayer: receiveIncomingVoiceModem %d\n", details);
	    _network_state == NETWORK_STATE_INCALL_WAIT;
	    if (_applicationLayer)
		_applicationLayer->receiveIncomingVoiceModem(address, details, buf + alen + 2, bufLen - alen - 2);
	}
    }
}

void Map27NetworkLayer::receiveProgress(uint8_t* buf, uint16_t bufLen)
{
    // 5.2.2.4.6
    if (bufLen >= 6)
    {
	uint8_t type = buf[0];
	Map27Address address;
	uint8_t alen = address.decode(buf + 1, bufLen - 1);
	if (bufLen >= alen + 2)
	{
	    uint8_t cause = buf[alen+1];
	    DEBUG_PRINTF("NetworkLayer: receiveProgress %d\n", cause);
	    if (_network_state == NETWORK_STATE_INCALL_WAIT)
	    {
		if (type == MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_PROGRESS_POSITIVE)
		    _network_state == NETWORK_STATE_ACTIVE;
		if (type == MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_PROGRESS_NEGATIVE)
		    _network_state == NETWORK_STATE_IDLE;
	    }
	    if (_applicationLayer)
		_applicationLayer->receiveProgress(address, cause);
	}
    }
}

void Map27NetworkLayer::receiveModemData(uint8_t* buf, uint16_t bufLen)
{
    // 5.2.2.5.2
    if (bufLen >= 1)
    {
	DEBUG_PRINTF("NetworkLayer: receiveModemData\n");
	if (_applicationLayer)
	    _applicationLayer->receiveModemData(buf + 1, bufLen - 1);
    }
}

void Map27NetworkLayer::receiveRadioPersonality(uint8_t* buf, uint16_t bufLen)
{
    DEBUG_PRINTF("receiveRadioPersonality\n");
    // 5.2.2.8.2
    RadioPersonality* msg = (RadioPersonality*)buf;
    if (bufLen == sizeof(RadioPersonality))
    {
	Map27Address address;
	address.decodeBasic(buf+1, bufLen - 1);
	uint8_t model = msg->serial_a >> 5;
	uint32_t serial = ((msg->serial_a & 0x03) << 16) | (msg->serial_b << 8) | msg->serial_c;
	
	DEBUG_PRINTF("NetworkLayer: receiveRadioPersonality: %d, %d, %02x, %02x, %02x, %02x\n", msg->manufacturers_code, model, serial, msg->facilities_a, msg->facilities_b, msg->facilities_c, msg->codings);
	if (_applicationLayer)
	    _applicationLayer->receiveRadioPersonality(address, msg->manufacturers_code, model, serial, 
						       msg->facilities_a, msg->facilities_b, msg->facilities_c, 
						       msg->codings);
    }
}

void Map27NetworkLayer::receiveProtocolInfo(uint8_t* buf, uint16_t bufLen)
{
    DEBUG_PRINTF("receiveProtocolInfo\n");
    // 5.2.2.8.9
    MsgProtocolInfo* msg = (MsgProtocolInfo*)buf;

    if (bufLen == sizeof(MsgProtocolInfo))
    {
	if (_network_state != NETWORK_STATE_ACTIVE)
	{
	    DEBUG_PRINTF("NetworkLayer: receiveProtocolInfo: %d. IDLE\n", msg->reason);
	    _network_state = NETWORK_STATE_IDLE;
	    _request_timer.stop();
	    if (_applicationLayer)
		_applicationLayer->receiveProtocolInfo(msg->reason);
	}
    }
}

void Map27NetworkLayer::receiveNumberingInfo(uint8_t* buf, uint16_t bufLen)
{
    DEBUG_PRINTF("receiveNumberingInfo\n");
    // 5.2.2.8.3
    MsgNumberingInfo* msg = (MsgNumberingInfo*)buf;

    if (bufLen >= sizeof(MsgNumberingInfo));
    {
	Map27Address address;
	address.decodeBasic(buf + 1, bufLen - 1);
	uint16_t individual_base = (msg->individual_base_a << 8) | msg->individual_base_b;
	bool     individual_base_3digit = individual_base & 0x8000;
	individual_base &= 0x7fff;
	uint16_t highest_ident = (msg->highest_ident_a << 8) | msg->highest_ident_b;
	uint16_t group_base = (msg->group_base_a << 8) | msg->group_base_b;
	bool     group_base_3digit = group_base & 0x8000;
	group_base &= 0x7fff;
	uint16_t highest_group = (msg->highest_group_a << 8) | msg->highest_group_b;
	// REVISIT: collect group addresses
	uint8_t i;
	for (i = 0; (sizeof(MsgNumberingInfo) + ((i+1) * sizeof(GroupAddress))) < bufLen; i++)
	{
	}
	DEBUG_PRINTF("NetworkLayer: receiveNumberingInfo: %d %d %d %d %d %d\n", 
		     individual_base, individual_base_3digit, highest_ident, 
		     group_base, group_base_3digit, highest_group); 
	// No state change
	if (_applicationLayer)
	    _applicationLayer->receiveNumberingInfo(address, 
						    individual_base, individual_base_3digit, highest_ident, 
						    group_base, group_base_3digit, highest_group);
    }
}

void Map27NetworkLayer::receiveRadioSettings(uint8_t* buf, uint16_t bufLen)
{
    DEBUG_PRINTF("receiveRadioSettings\n");
    // 5.2.2.8.7
    MsgRadioSettings* msg = (MsgRadioSettings*)buf;

    // CAUTION: Tait 2040 does not send codings!!! Urk
    if (bufLen >= sizeof(MsgRadioSettings) - 1);
    {
	DEBUG_PRINTF("NetworkLayer: receiveRadioSettings: %02x %02x %02x %02x\n", msg->controls_a, msg->controls_b, 
		     msg->controls_c, msg->codings);
	// No state change
	if (_applicationLayer)
	    _applicationLayer->receiveRadioSettings(msg->controls_a, msg->controls_b, 
						    msg->controls_c, msg->codings);
    }
}

void Map27NetworkLayer::receiveOperatingCondition(uint8_t* buf, uint16_t bufLen)
{
    DEBUG_PRINTF("receiveOperatingCondition\n");
    // 5.2.2.8.5
    MsgOperatingCondition* msg = (MsgOperatingCondition*)buf;

    if (bufLen == sizeof(MsgOperatingCondition));
    {
	DEBUG_PRINTF("NetworkLayer: receiveOperatingCondition: %02x %d %02x\n", msg->conditions, 
		     msg->field_strength, msg->max_call_duration);
	// No state change
	if (_applicationLayer)
	    _applicationLayer->receiveOperatingCondition(msg->conditions, msg->field_strength, msg->max_call_duration);
    }
}
void Map27NetworkLayer::receiveNetworkInformation(uint8_t* buf, uint16_t bufLen)
{
    DEBUG_PRINTF("receiveNetworkInformation\n");
    // 5.2.2.8.5
    MsgNetworkInformation* msg = (MsgNetworkInformation*)buf;

    if (bufLen == sizeof(MsgNetworkInformation));
    {
	uint16_t channel = (msg->channel_a & 0x3) | msg->channel_b;
	uint16_t sys_ident = (msg->sys_ident_a & 0x7f) | msg->sys_ident_b;
	DEBUG_PRINTF("NetworkLayer: receiveNetworkInformation: %d %d\n", channel, sys_ident);
	// No state change
	if (_applicationLayer)
	    _applicationLayer->receiveNetworkInformation(channel, sys_ident);
    }
}

void Map27NetworkLayer::receiveRadioTestResult(uint8_t* buf, uint16_t bufLen)
{
    // 5.2.2.4.8.13
    if (bufLen >= 1)
    {
	DEBUG_PRINTF("NetworkLayer: receiveRadioTestResult\n");
	if (_applicationLayer)
	    _applicationLayer->receiveRadioTestResult(buf + 1, bufLen - 1);
    }
}

void Map27NetworkLayer::output_application_error(uint8_t type)
{
    if (_applicationLayer)
	_applicationLayer->error(type);
}
