// NetworkLayer.h
//
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2013 Mike McCauley
// $Id: NetworkLayer.h,v 1.6 2013/07/14 06:32:28 mikem Exp $


#ifndef MAP27_NETWORKLAYER_H
#define MAP27_NETWORKLAYER_H

#include <stdint.h>

#include "Timer.h"

// Timeout for pending Netwrok Layer requests in msec
#define MAP27_NETWORK_REQUEST_TIMEOUT 10000

// Largest network layer message
#define MAP27_NETWORK_MESSAGE_MAX_MESSAGE_LEN 43

// From Table 5-1
// Note that bother send and receive message numbers are here, and in some cases the same number
// appears twide, one for D->R and one for R->D
#define MAP27_NETWORK_MESSAGE_TYPE_SEND_STATUS                    0x80
#define MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_STATUS                 0x80
#define MAP27_NETWORK_MESSAGE_TYPE_ACK_POSITIVE                   0xc0
#define MAP27_NETWORK_MESSAGE_TYPE_ACK_QUEUEING                   0xd0
#define MAP27_NETWORK_MESSAGE_TYPE_ACK_NEGATIVE                   0xe0
#define MAP27_NETWORK_MESSAGE_TYPE_SEND_SST                       0x81
#define MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_SST                    0x81
#define MAP27_NETWORK_MESSAGE_TYPE_SEND_MST                       0x82
#define MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_MST                    0x82
#define MAP27_NETWORK_MESSAGE_TYPE_SETUP_VOICE_MODEM              0xa4
#define MAP27_NETWORK_MESSAGE_TYPE_SETUP_EMERGENCY_VOICE_MODEM    0xa5
#define MAP27_NETWORK_MESSAGE_TYPE_SETUP_PROGRESS_POSITIVE        0xc4
#define MAP27_NETWORK_MESSAGE_TYPE_SETUP_PROGRESS_QUEUEING        0xd4
#define MAP27_NETWORK_MESSAGE_TYPE_SETUP_PROGRESS_NEGATIVE        0xe4
#define MAP27_NETWORK_MESSAGE_TYPE_INCOMING_VOICE_MODEM           0xa4
#define MAP27_NETWORK_MESSAGE_TYPE_INCOMING_EMERGENCY_VOICE_MODEM 0xa5
#define MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_PROGRESS_POSITIVE      0xc5
#define MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_PROGRESS_WARNING       0xd5
#define MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_PROGRESS_NEGATIVE      0xe5
#define MAP27_NETWORK_MESSAGE_TYPE_SEND_MODEM_DATA                0xa3
#define MAP27_NETWORK_MESSAGE_TYPE_RECEIVE_MODEM_DATA             0xa3
#define MAP27_NETWORK_MESSAGE_TYPE_DISCONNECT_NORMAL              0x86
#define MAP27_NETWORK_MESSAGE_TYPE_DISCONNECT_CANCEL              0xa6
#define MAP27_NETWORK_MESSAGE_TYPE_CLEARED_NORMAL                 0x86
#define MAP27_NETWORK_MESSAGE_TYPE_CLEARED_ABNORMAL               0xa6
#define MAP27_NETWORK_MESSAGE_TYPE_DIVERSION_REQUEST              0x87
#define MAP27_NETWORK_MESSAGE_TYPE_DIVERSION_CANCEL               0xa7
#define MAP27_NETWORK_MESSAGE_TYPE_DIVERSION_ACK_POSITIVE         0xc7
#define MAP27_NETWORK_MESSAGE_TYPE_DIVERSION_ACK_NEGATIVE         0xe7
#define MAP27_NETWORK_MESSAGE_TYPE_RADIO_INTERROGATION            0xb0
#define MAP27_NETWORK_MESSAGE_TYPE_RADIO_PERSONALITY              0xb0
#define MAP27_NETWORK_MESSAGE_TYPE_NUMBERING_INFORMATION          0xb1
#define MAP27_NETWORK_MESSAGE_TYPE_RADIO_CONTROL                  0xb2
#define MAP27_NETWORK_MESSAGE_TYPE_OPERATING_CONDITION            0xb2
#define MAP27_NETWORK_MESSAGE_TYPE_NETWORK_INFORMATION            0xb5
#define MAP27_NETWORK_MESSAGE_TYPE_RADIO_MANAGEMENT               0xb3
#define MAP27_NETWORK_MESSAGE_TYPE_RADIO_SETTINGS                 0xb3
#define MAP27_NETWORK_MESSAGE_TYPE_PROTOCOL_INFO                  0xb4
#define MAP27_NETWORK_MESSAGE_TYPE_VOLUME_CONTROL                 0xb6
#define MAP27_NETWORK_MESSAGE_TYPE_DIALLED_STRING                 0xb7
#define MAP27_NETWORK_MESSAGE_TYPE_RADIO_TEST                     0x01
#define MAP27_NETWORK_MESSAGE_TYPE_RADIO_TEST_RESULT              0x01
#define MAP27_NETWORK_MESSAGE_TYPE_RESERVED                       0x00

// Status 5.2.2.1.1, 5.2.2.1.2
// Values 0..31 are permitted
// Need to see MPT1327 for these, section 13
#define MAP27_NETWORK_MESSAGE_STATUS_OFF_HOOK                     0x00
// The user defined messages 00001 to 11110 are not given here
#define MAP27_NETWORK_MESSAGE_STATUS_ON_HOOK                      0x1f

// Status Ack cause 5.2.2.1.3
// For type 0xc0
#define MAP27_NETWORK_MESSAGE_STATUS_CAUSE_SUCCESS                0x00
// For type 0xd0
#define MAP27_NETWORK_MESSAGE_STATUS_CAUSE_BUSY_WAIT              0x02
#define MAP27_NETWORK_MESSAGE_STATUS_CAUSE_ENGAGED_WAIT           0x0a
#define MAP27_NETWORK_MESSAGE_STATUS_CAUSE_DIVERTING_WAIT         0x26
// For type 0xe0
#define MAP27_NETWORK_MESSAGE_STATUS_CAUSE_ABORTED                0x08
#define MAP27_NETWORK_MESSAGE_STATUS_CAUSE_INVALID                0x03
#define MAP27_NETWORK_MESSAGE_STATUS_CAUSE_OVERLOAD               0x0b
#define MAP27_NETWORK_MESSAGE_STATUS_CAUSE_ABANDONED              0x04
#define MAP27_NETWORK_MESSAGE_STATUS_CAUSE_ENGAGED                0x0c
#define MAP27_NETWORK_MESSAGE_STATUS_CAUSE_DIVERTED               0x06
#define MAP27_NETWORK_MESSAGE_STATUS_CAUSE_DIVERTED_GROUP         0x16
#define MAP27_NETWORK_MESSAGE_STATUS_CAUSE_DIVERTED_NOT_AVAILABLE 0x07

// For Send SST 5.2.2.2.1
#define MAP27_NETWORK_MESSAGE_SEND_SST_CODING_MASK                0xf0
#define MAP27_NETWORK_MESSAGE_SEND_SST_CODING_FREE_FORMAT         0x00
#define MAP27_NETWORK_MESSAGE_SEND_SST_CODING_BCD_RADIO_PATH      0x10
#define MAP27_NETWORK_MESSAGE_SEND_SST_CODING_CCITT_RADIO_PATH    0x20
#define MAP27_NETWORK_MESSAGE_SEND_SST_CODING_BINARY              0x80
#define MAP27_NETWORK_MESSAGE_SEND_SST_CODING_BCD                 0x90
#define MAP27_NETWORK_MESSAGE_SEND_SST_CODING_CCITT2              0xa0
#define MAP27_NETWORK_MESSAGE_SEND_SST_CODING_ASCII               0xb0
#define MAP27_NETWORK_MESSAGE_SEND_SST_CODING_OCTETS              0xf0

// For SST ACK 5.2.2.2.3 and MST ACK 5.2.2.3.3
// For type 0xc0
#define MAP27_NETWORK_MESSAGE_SST_ACK_SUCCESS                     0x00
// For type 0xd0
#define MAP27_NETWORK_MESSAGE_SST_ACK_BUSY_WAIT                   0x02
#define MAP27_NETWORK_MESSAGE_SST_ACK_ENGAGED_WAIT                0x0a
#define MAP27_NETWORK_MESSAGE_SST_ACK_DIVERTED_WAIT               0x26
#define MAP27_NETWORK_MESSAGE_MST_ACK_MST_NOT_SUPPORTED_SST       0x22
/// For type 0xe0
#define MAP27_NETWORK_MESSAGE_SST_ACK_ABORTED                     0x08
#define MAP27_NETWORK_MESSAGE_SST_ACK_INVALID                     0x03
#define MAP27_NETWORK_MESSAGE_SST_ACK_OVERLOAD                    0x0b
#define MAP27_NETWORK_MESSAGE_SST_ACK_ABANDONED                   0x04
#define MAP27_NETWORK_MESSAGE_SST_ACK_ENGAGED                     0x0c
#define MAP27_NETWORK_MESSAGE_SST_ACK_DIVERTED                    0x06
#define MAP27_NETWORK_MESSAGE_SST_ACK_DIVERTED_GROUP              0x16
#define MAP27_NETWORK_MESSAGE_SST_ACK_DIVERTED_NOT_AVAILABLE      0x07
#define MAP27_NETWORK_MESSAGE_MST_ACK_MST_NOT_SUPPORTED_ABORTED   0x0d

// For Setup Voice/Modem 5.2.2.4.1
// Also for Setup Emergency Voice/Modem 5.2.2.4.2
#define MAP27_NETWORK_MESSAGE_SETUP_VOICE_MODEM_DETAILS_INCLUDE          0x40
#define MAP27_NETWORK_MESSAGE_SETUP_VOICE_MODEM_DETAILS_GROUP            0x20
#define MAP27_NETWORK_MESSAGE_SETUP_VOICE_MODEM_DETAILS_MODEM            0x10
#define MAP27_NETWORK_MESSAGE_SETUP_VOICE_MODEM_DETAILS_PRIORITY         0x08
#define MAP27_NETWORK_MESSAGE_SETUP_VOICE_MODEM_DETAILS_SPECIAL          0x01

// For Setup Progress 5.2.2.4.3
// For type 0xc4
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_CALL_CONNECTED         0x00
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_INCLUDE_CALL_CONNECTED 0x00
// For type 0xd4
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_ALERTING               0x01
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_BUSY                   0x02
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_ENGAGED_WAIT           0x0a
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_EMERGENCY_PROCEEDING   0x05
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_DIVERTED_WAIT          0x26
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_DIVERTED_GROUP_WAIT    0x36
// For type 0xe4
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_ABORTED                0x08
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_INVALID                0x03
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_OVERLOAD               0x0b
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_ABANDONED              0x04
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_ENGAGED                0x0c
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_DIVERTED               0x06
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_DIVERTED_GROUP         0x16
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_DIVERTED_NOT_AVAILABLE 0x0e
#define MAP27_NETWORK_MESSAGE_SETUP_PROGRESS_CAUSE_CALLBACK_ACCEPTED      0x07

// For Incoming Voice/Modem 5.2.2.4.4
// and Incoming Emergency Voice/Modem 5.2.2.4.5
#define MAP27_NETWORK_MESSAGE_INCOMING_VOICE_MODEM_DETAILS_INCLUDE          0x40
#define MAP27_NETWORK_MESSAGE_INCOMING_VOICE_MODEM_DETAILS_GROUP            0x20
#define MAP27_NETWORK_MESSAGE_INCOMING_VOICE_MODEM_DETAILS_DATA             0x10
#define MAP27_NETWORK_MESSAGE_INCOMING_VOICE_MODEM_DETAILS_HOOK_REQUIRED    0x08
#define MAP27_NETWORK_MESSAGE_INCOMING_VOICE_MODEM_DETAILS_CUSTOM_SERVICE_1 0x05
#define MAP27_NETWORK_MESSAGE_INCOMING_VOICE_MODEM_DETAILS_CUSTOM_SERVICE_2 0x06
#define MAP27_NETWORK_MESSAGE_INCOMING_VOICE_MODEM_DETAILS_CUSTOM_SERVICE_3 0x07

// For Receive Progress 5.2.2.4.6
// For type 0xc5
#define MAP27_NETWORK_MESSAGE_RECEIVE_PRGRESS_CAUSE_CONNECTED               0x00
// For type 0xd5
#define MAP27_NETWORK_MESSAGE_RECEIVE_PRGRESS_CAUSE_TRANSMISSION_DISABLED   0x01
#define MAP27_NETWORK_MESSAGE_RECEIVE_PRGRESS_CAUSE_BUSY_WAIT               0x02
#define MAP27_NETWORK_MESSAGE_RECEIVE_PRGRESS_CAUSE_CLEAR_DOWN_TIMER        0x0a
// For type 0xe5
#define MAP27_NETWORK_MESSAGE_RECEIVE_PRGRESS_CAUSE_REJECTED                0x03
#define MAP27_NETWORK_MESSAGE_RECEIVE_PRGRESS_CAUSE_CALLBACK                0x07

// For Disconnect 5.2.2.6.1
// For type 0x86
#define MAP27_NETWORK_MESSAGE_DISCONNECT_CAUSE_NORMAL                       0x07
// For type 0xa6
#define MAP27_NETWORK_MESSAGE_DISCONNECT_CAUSE_CANCEL_TRANS                 0x08
#define MAP27_NETWORK_MESSAGE_DISCONNECT_CAUSE_CANCEL_INCLUDE               0x09
#define MAP27_NETWORK_MESSAGE_DISCONNECT_CAUSE_ABORT_DIVERSION              0x0a
#define MAP27_NETWORK_MESSAGE_DISCONNECT_CAUSE_CANCEL_DATA_SETUP            0x0c

// For Cleared 5.2.2.6.2
// For type 0x86
#define MAP27_NETWORK_MESSAGE_CLEARED_CAUSE_NORMAL                          0x0b
// For type 0xa6
#define MAP27_NETWORK_MESSAGE_CLEARED_CAUSE_UNSPECIFIED                     0x00
#define MAP27_NETWORK_MESSAGE_CLEARED_CAUSE_RADIO_GENERATED                 0x01
#define MAP27_NETWORK_MESSAGE_CLEARED_CAUSE_NOT_AVAILABLE                   0x04
#define MAP27_NETWORK_MESSAGE_CLEARED_CAUSE_TOO_LONG                        0x05
#define MAP27_NETWORK_MESSAGE_CLEARED_CAUSE_CODING_NOT_POSSIBLE             0x06
#define MAP27_NETWORK_MESSAGE_CLEARED_CAUSE_ABNORMAL                        0x0e

// For Diversion Request Info 5.2.2.7.1
#define MAP27_NETWORK_MESSAGE_DIVERSION_REQUEST_INFO_SPEECH_DATA            0x00
#define MAP27_NETWORK_MESSAGE_DIVERSION_REQUEST_INFO_SPEECH                 0x08
#define MAP27_NETWORK_MESSAGE_DIVERSION_REQUEST_INFO_DATA                   0x10
#define MAP27_NETWORK_MESSAGE_DIVERSION_REQUEST_INFO_THIRD_PARTY            0x01

// For Diversion Cancel Info 5.2.2.7.2
#define MAP27_NETWORK_MESSAGE_DIVERSION_CANCEL_INFO_SPEECH_DATA            0x00
#define MAP27_NETWORK_MESSAGE_DIVERSION_CANCEL_INFO_SPEECH                 0x08
#define MAP27_NETWORK_MESSAGE_DIVERSION_CANCEL_INFO_DATA                   0x10
#define MAP27_NETWORK_MESSAGE_DIVERSION_CANCEL_INFO_SELF                   0x00
#define MAP27_NETWORK_MESSAGE_DIVERSION_CANCEL_INFO_THIRD_PARTY            0x01
#define MAP27_NETWORK_MESSAGE_DIVERSION_CANCEL_INFO_RECIPIENT              0x02

// For Radio Interrogation 5.2.2.8.1
#define MAP27_NETWORK_MESSAGE_RADIO_INTERROGATION_PERSONALITY     0x00
#define MAP27_NETWORK_MESSAGE_RADIO_INTERROGATION_NUMBERING       0x01
#define MAP27_NETWORK_MESSAGE_RADIO_INTERROGATION_STATUS          0x02
#define MAP27_NETWORK_MESSAGE_RADIO_INTERROGATION_OPERATING_CONDITION 0x03
#define MAP27_NETWORK_MESSAGE_RADIO_INTERROGATION_NETWORK_INFORMATION 0x04
// SPARE and Reserved not defined here

// For RADIO PERSONALITY 5.2.2.8.2
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_FACILITIESA_VOICE          0x80
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_FACILITIESA_MODEM          0x40
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_FACILITIESA_STATUS         0x20
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_FACILITIESA_SST            0x10
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_FACILITIESA_MST            0x08
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_FACILITIESA_DIVERSION      0x04
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_FACILITIESA_LOGGING        0x02
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_FACILITIESA_SHORT_DATA     0x01

#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_CODINGS_BCD_RADIO_PATH     0x80
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_CODINGS_CCITT_RADIO_PATH   0x40
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_CODINGS_BINARY             0x20
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_CODINGS_BCD                0x10
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_CODINGS_CCITT2             0x08
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_CODINGS_ASCII              0x04
#define MAP27_NETWORK_MESSAGE_RADIO_PERSONALITY_CODINGS_OCTETS             0x02

// For Protocol Info 5.2.2.8.9
#define MAP27_NETWORK_MESSAGE_PROTOCOL_INFO_UNRECOGNISED         0x01
#define MAP27_NETWORK_MESSAGE_PROTOCOL_INFO_NOT_SUPPORTED        0x02
#define MAP27_NETWORK_MESSAGE_PROTOCOL_INFO_STATE_MISMATCH       0x03
#define MAP27_NETWORK_MESSAGE_PROTOCOL_INFO_CODING_NOT_SUPPORTED 0x04
#define MAP27_NETWORK_MESSAGE_PROTOCOL_INFO_BUSY                 0x05
#define MAP27_NETWORK_MESSAGE_PROTOCOL_INFO_READY                0x06
// SPARE and RESERVED not given here

// For Radio Management 5.2.2.8.7 and Radio Settings 5.2.2.8.8
// Masks for controls_a
#define MAP27_NETWORK_MESSAGE_RADIO_MANAGEMENT_CONTROLS_VOICE    0x80
#define MAP27_NETWORK_MESSAGE_RADIO_MANAGEMENT_CONTROLS_MODEM    0x40
#define MAP27_NETWORK_MESSAGE_RADIO_MANAGEMENT_CONTROLS_STATUS   0x20
#define MAP27_NETWORK_MESSAGE_RADIO_MANAGEMENT_CONTROLS_SST      0x10
#define MAP27_NETWORK_MESSAGE_RADIO_MANAGEMENT_CONTROLS_MST      0x08
#define MAP27_NETWORK_MESSAGE_RADIO_MANAGEMENT_CONTROLS_DIVERSION 0x04
#define MAP27_NETWORK_MESSAGE_RADIO_MANAGEMENT_CONTROLS_LOGGING  0x02

// For Operating Condition 5.2.2.8.5
#define MAP27_NETWORK_MESSAGE_OPERATING_CONDITION_CONDITION_IN_CONTACT 0x04
#define MAP27_NETWORK_MESSAGE_OPERATING_CONDITION_CONDITION_OFF_HOOK   0x02
#define MAP27_NETWORK_MESSAGE_OPERATING_CONDITION_CONDITION_TRANSMITTING 0x01
#define MAP27_NETWORK_MESSAGE_OPERATING_CONDITION_FIELD_STRENGTH_DUMMY 0x00
#define MAP27_NETWORK_MESSAGE_OPERATING_CONDITION_CALL_DURATION_RESERVED 0x00
#define MAP27_NETWORK_MESSAGE_OPERATING_CONDITION_CALL_DURATION_INFINITE 0xff

// For Volume Control 5.2.2.8.10
#define MAP27_NETWORK_MESSAGE_VOLUME_CONTROL_POINT_MODEM                 0x04
#define MAP27_NETWORK_MESSAGE_VOLUME_CONTROL_POINT_ALERTS                0x02
#define MAP27_NETWORK_MESSAGE_VOLUME_CONTROL_POINT_SPEECH                0x01
#define MAP27_NETWORK_MESSAGE_VOLUME_SET_UP                              0x01
#define MAP27_NETWORK_MESSAGE_VOLUME_SET_DOWN                            0x02
#define MAP27_NETWORK_MESSAGE_VOLUME_SET_VALUE                           0x03
#define MAP27_NETWORK_MESSAGE_VOLUME_SET_PRESET                          0x04
#define MAP27_NETWORK_MESSAGE_VOLUME_SET_MANUAL                          0x05

class Map27Address;
class Map27DataLayer;
class Map27ApplicationLayer;

/////////////////////////////////////////////////////////////////////
/// \class Map27NetworkLayer NetworkLayer.h <Map27/NetworkLayer.h>
/// \brief Manage the Network Layer of the Map27 protocol stack
///
/// This class receives messages to be sent from the Map27ApplicationLayer and formats them
/// for transmission to the Radio by Map27DataLayer. It also received messages from the Radio received by 
/// Map27DataLayer and delivers them to Map27ApplicationLayer.
class Map27NetworkLayer
{
public:

    /// Network Layer state _network_state values.
    /// The network layer is always in exactly one of these states
    typedef enum
    {
	NETWORK_STATE_IDLE = 1,             ///< Idle, ready to start a new transaction 
	NETWORK_STATE_OUTCALL_WAIT,         ///< Waiting for an outgoing voice/data call to be set up
	NETWORK_STATE_OUTCALL_INCLUDE_WAIT, ///< Waiting for an additional outgoing voice/data call to be set up
	NETWORK_STATE_ACTIVE,               ///< Voice/data call is active
	NETWORK_STATE_INCALL_WAIT,          ///< Waiting for an incoming voice/data call to be set up
	NETWORK_STATE_STATUS_WAIT,          ///< Waiting for status send to complete
	NETWORK_STATE_SST_WAIT,             ///< Waiting for an SST data send to complete
	NETWORK_STATE_MST_WAIT,             ///< Waiting for an MST data send to complete
	NETWORK_STATE_DIVERSION_WAIT        ///< Waiting for a diversion request to complete
    } NetworkState;

    /// Constructor
    Map27NetworkLayer();

    /// Sets the pointer to the DataLayer to use
    /// to communicate with the Radio
    /// \param[in] dataLayer Pointer to the Data Layer
    void setDataLayer(Map27DataLayer* dataLayer);

    /// Get the current DataLayer
    /// \return The current Data Layer pointer, or NULL if previously set.
    Map27DataLayer* dataLayer();

    /// Sets the pointer to the ApplicationLayer to use
    /// to deliver messages to the application
    /// \param[in] applicationLayer Pointer to the Application Layer
    void setApplicationLayer(Map27ApplicationLayer* applicationLayer);

    /// Get the current Application Layer
    /// \return The curent Application Layer pointer, or NULL if previously set.
    Map27ApplicationLayer* applicationLayer();

    /// The current state of the Network Layer state machine
    /// \return The current state
    NetworkState state();

    /// Whether the Network Layer state machine is idle
    /// \return true if the current Network Layer state mahcine state is IDLE
    bool is_idle();

    /// Cause the entire protocol stack to 
    /// power up and intialise to its startup state. Network state is IDLE.
    void power_on();

    /// Resets the Network Layer to 
    /// its power_on state
    void reset();

    /// Polls Network Layer timer
    /// and lower layers of the protocol stack for activity.
    /// In Unix/Linux and Arduino environments, you are required to call this at least as 
    /// frequently as octets might be received from the radio. At least every 10msec would 
    /// be appropriate for a 9600 baud connection.
    /// Note: Map27ApplicationLayer does this automatically when its poll() function is called.
    void poll();

    /// \name Inputs from Data Layer:
    /// These events are sent by Map27DataLAyer in repsonse to events in lower layers.
    /// You might override them in a custom subclass.

    /// Called by Data Layer
    /// when the data link becomes READY
    virtual void link_ready();

    /// Called by Data Layer
    /// when the data link stops being READY
    virtual void link_failure();

    /// Called by Data Layer
    /// when a packet is accepted by the Data Layer for transmission
    virtual void packet_accepted();

    /// Called by Data Layer
    /// when a packet is rejected by the DataLayer for transmission
    virtual void packet_rejected();

    /// Called by Data Layer
    /// when a message is received from the radio
    /// \param[in] buf Contents of the message
    /// \param[in] bufLen Number of bytes in buf
    virtual void network_layer_packet(uint8_t* buf, uint16_t bufLen);
	
    /// \name Inputs from ApplicationLayer
    /// Applications can send messages by calling these members

    /// Senda SEND STATUS 5.2.2.1.1 message
    /// to a specified radio in the network.
    /// \param[in] address Called party address
    /// \param[in] status Status number. One of MAP27_NETWORK_MESSAGE_STATUS_*
    /// \return true if the message was queued for transmission by Data Layer
    bool sendStatus(const Map27Address& address, uint8_t status);

    /// Sends a SEND SST 5.2.2.2.1 message
    /// to a specified radio in the network.
    /// \param[in] address Called party address
    /// \param[in] buf Data to send
    /// \param[in] bufLen Number of bytes in buf
    /// \return true if the message was queued for transmission by Data Layer
    bool sendSST(const Map27Address& address, uint8_t* buf, uint16_t bufLen);

    /// Sends a SEND MST 5.2.2.3.1 message
    /// to a specified radio in the network.
    /// \param[in] address Called party address
    /// \param[in] buf Data to send
    /// \param[in] bufLen Number of bytes in buf
    /// \return true if the message was queued for transmission by Data Layer
    bool sendMST(const Map27Address& address, uint8_t* buf, uint16_t bufLen);

    /// Sends a SETUP VOICE / SETUP MODEM 5.2.2.4.1 message
    /// to a specified radio in the network.
    /// \param[in] address Called party address
    /// \param[in] details Call details One of MAP27_NETWORK_MESSAGE_SETUP_VOICE_MODEM_DETAILS_*.
    /// \return true if the message was queued for transmission by Data Layer
    bool setupVoiceModem(const Map27Address& address, uint8_t details);

    /// Sends a SETUP EMERGENCY VOICE / SETUP EMERGENCY MODEM 5.2.2.4.2 message
    /// to a specified radio in the network.
    /// \param[in] address Called party address
    /// \param[in] details Call details One of MAP27_NETWORK_MESSAGE_SETUP_VOICE_MODEM_DETAILS_*.
    /// \return true if the message was queued for transmission by Data Layer
    bool setupEmergencyVoiceModem(const Map27Address& address, uint8_t details);

    /// Sends a SEND MODEM DATA 5.2.2.5.1 message
    /// to a specified radio in the network.
    /// \param[in] buf Data to send
    /// \param[in] bufLen Number of bytes in buf
    /// \return true if the message was queued for transmission by Data Layer
    bool sendModemData(uint8_t* buf, uint16_t bufLen);

    /// Sends a DISCONNECT 5.2.2.6.1 message
    /// to a specified radio in the network.
    /// \param[in] address Called or connected party address
    /// \param[in] cause Cancellation or disconnection reason. One of MAP27_NETWORK_MESSAGE_DISCONNECT_CAUSE_*
    /// \return true if the message was queued for transmission by Data Layer
    bool sendDisconnect(const Map27Address& address, uint8_t cause);

    /// Sends a CLEARED 5.2.2.6.2 message
    /// to a specified radio in the network.
    /// \param[in] address Called, calling or connected party address
    /// \param[in] cause Cancellation or disconnection reason. One of  MAP27_NETWORK_MESSAGE_CLEARED_CAUSE_*
    /// \return true if the message was queued for transmission by Data Layer
    bool sendCleared(const Map27Address& address, uint8_t cause);

    /// Sends a DIVERSION REQUEST 5.2.2.7.1 message
    /// to divert messages to another unit
    /// \param[in] address Address of he party to which calls are diverted
    /// \param[in] info Informationof the diversion. One of MAP27_NETWORK_MESSAGE_DIVERSION_REQUEST_INFO_*
    /// \param[in] diversionAddress Bloked party address in a third party diversion
    /// \return true if the message was queued for transmission by Data Layer
    bool sendDiversionRequest(const Map27Address& address, uint8_t info, const Map27Address& diversionAddress);

    /// Sends a DIVERSION CANCEL 5.2.2.7.2 message
    /// to cancel a previous diversion.
    /// \param[in] address Address of the party whose calls are no longer diverted
    /// \param[in] info Information of the diversion. One of MAP27_NETWORK_MESSAGE_DIVERSION_CANCEL_*
    /// \return true if the message was queued for transmission by Data Layer
    bool sendDiversionCancel(const Map27Address& address, uint8_t info);

    /// Sends a RADIO INTERROGATION message
    /// to the radio, requesting informaiton about its state or configuration. Radio will
    /// respond some time later with a reply message.
    /// \param[in] reason Point of interrogation. One of MAP27_NETWORK_MESSAGE_RADIO_INTERROGATION_*
    /// \return true if the message was queued for transmission by Data Layer
    bool sendRadioInterrogation(uint8_t reason);

    /// Sends a VOLUME CONTROL 5.2.2.8.10 message
    /// to the radio.
    /// \param[in] control_port Address of the volume control port. One of MAP27_NETWORK_MESSAGE_VOLUME_CONTROL_POINT_*
    /// \param[in] volume Volume set. One of MAP27_NETWORK_MESSAGE_VOLUME_SET_* and others.
    /// \return true if the message was queued for transmission by Data Layer
    bool sendVolumeControl(uint8_t control_port, uint8_t volume);

    /// Sends a DIALLED STRING 5.2.2.8.11 message
    /// to the radio.
    /// \param[in] buf The dialled string
    /// \param[in] bufLen The number of bytes in the dialled digits
    /// \return true if the message was queued for transmission by Data Layer
    bool sendDialledString(uint8_t* buf, uint16_t bufLen);

    /// Sends a RADIO TESET 5.2.2.8.12 message
    /// to the radio.
    /// \param[in] buf Test controls. Depends on radio type
    /// \param[in] bufLen Number of bytes in buf.
    /// \return true if the message was queued for transmission by Data Layer
    bool sendRadioTest(uint8_t* buf, uint16_t bufLen);

protected:
    /// Called internally when a message is recieved by the Data LAyer
    /// and requires decoding and dispatching.
    /// \param[in] buf Data in the message
    /// \param[in] bufLen Number of bytes in buf
    void decode_network_layer_packet(uint8_t* buf, uint16_t bufLen);

    /// Called internally when a pending request times out.
    /// Cancels the reques and reurns the Network Layer to IDLE state
    void request_timeout();

    /// \name Message decoders
    /// These functions are called when for different types of mesasge. They are passed
    /// a pointer to the raw message data buffer,and are repsonsibel for decoding the mesage
    /// and calling the appropriate member funciton in the Application Layer.
    /// You may want to subclass Map27NetworkLayer and override one or more of these functions
    /// for special behaviour.

    /// Decode a RECIEVE STATUS (5.2.2.1.2) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveStatus(uint8_t* buf, uint16_t bufLen);

    /// Decode a STATUS ACK (5.2.2.1.3) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveStatusAck(uint8_t* buf, uint16_t bufLen);

    /// Decode a RECEIVE SST (5.2.2.2.2) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveSST(uint8_t* buf, uint16_t bufLen);

    /// Decode a SST ACK (5.2.2.2.3) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveSSTAck(uint8_t* buf, uint16_t bufLen);

    /// Decode a RECEIVE MST (5.2.2.3.2) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveMST(uint8_t* buf, uint16_t bufLen);

    /// Decode a MST ACK (5.2.2.3.3) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveMSTAck(uint8_t* buf, uint16_t bufLen);

    /// Decode a SETUP PROGRESS (5.2.2.4.3) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveSetupProgress(uint8_t* buf, uint16_t bufLen);

    /// Decode a INCOMING VOICE CALL / INCOMING MODEM CALL (5.2.2.4.4) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveIncomingVoiceModem(uint8_t* buf, uint16_t bufLen);

    /// Decode a  message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveProgress(uint8_t* buf, uint16_t bufLen);

    /// Decode a RECEIVE PROGRESS (5.2.2.4.6) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveModemData(uint8_t* buf, uint16_t bufLen);

    /// Decode a CLEARED (5.2.2.6.2) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveCleared(uint8_t* buf, uint16_t bufLen);

    /// Decode a RADIO PERSONALITY (5.2.2.8.2) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveRadioPersonality(uint8_t* buf, uint16_t bufLen);

    /// Decode a  message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveProtocolInfo(uint8_t* buf, uint16_t bufLen);

    /// Decode a PROTOCOL INFO (5.2.2.8.9) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveNumberingInfo(uint8_t* buf, uint16_t bufLen);

    /// Decode a RADIO SETTINGS (5.2.2.8.8) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveRadioSettings(uint8_t* buf, uint16_t bufLen);

    /// Decode a OPERATING CONDITION (5.2.2.8.5) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveOperatingCondition(uint8_t* buf, uint16_t bufLen);

    /// Decode a NETWORK INFORMATION (5.2.2.8.6) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveNetworkInformation(uint8_t* buf, uint16_t bufLen);

    /// Decode a RADIO TEST RESULT (5.2.2.8.13) message
    /// and pass it to the appropriate ApplicationLAyer function
    /// \param[in] buf Pointer to the raw message data
    /// \param[in] bufLen Number of bytes in buf
    virtual void receiveRadioTestResult(uint8_t* buf, uint16_t bufLen);

    /// \name Outputs to Application Layer
    /// These functions are called to deliver events to the Data Layer
    /// You may want to override themfor special behaviour

    /// Tells the Application Layer about an error
    /// condition that was detected by the Network Layer
    /// \param[in] type The type of error. One of MAP27_APPLICATION_ERROR_*
    virtual void output_application_error(uint8_t type);

    // Callback glue
    static void request_timer_callback_glue(void* target);

private:
    // Definitions of various on-the-wire message formats
    // All these network layer message structs must be packed to 1 byte
#pragma pack(1)
    typedef struct
    {
	uint8_t  type;
	uint8_t  data[0];
    } GenericHeader;

    typedef struct
    {
	uint8_t  type;
	uint8_t  pfix;
	uint8_t  ident_a;
	uint8_t  ident_b;
	uint8_t  data[0];
    } IdentHeader;
    typedef struct
    {
	uint8_t  type;
	uint8_t  pfix;
	uint8_t  ident_a;
	uint8_t  ident_b;
	uint8_t  adesc;
	uint8_t  address[0]; // length indicated by previous octet
    } AddressHeader;

    // 5.2.2.1.1, 5.2.2.1.2
    typedef struct
    {
	AddressHeader header;
	uint8_t  status;
    } MsgStatus;

    // 5.2.2.1.3
    typedef struct
    {
	AddressHeader header;
	uint8_t  cause;
    } MsgStatusAck;

    // 5.2.2.6.2
    typedef struct
    {
	IdentHeader header;
	uint8_t  cause;
    } MsgCleared;

    // 5.2.2.8.9
    typedef struct
    {
	uint8_t  type;
	uint8_t  reason;
    } MsgProtocolInfo;

    // 5.2.2.8.1
    typedef struct
    {
	uint8_t  type;
	uint8_t  reason;
    } MsgRadioInterrogation;

    // 5.2.2.8.2
    typedef struct
    {
	uint8_t  type;
	uint8_t  pfix;
	uint8_t  ident_a;
	uint8_t  ident_b;
	uint8_t  manufacturers_code;
	uint8_t  serial_a;
	uint8_t  serial_b;
	uint8_t  serial_c;
	uint8_t  facilities_a;
	uint8_t  facilities_b;
	uint8_t  facilities_c;
	uint8_t  codings;
    } RadioPersonality;

    // 5.2.2.8.3
    typedef struct
    {
	uint8_t  pfix;
	uint8_t  ident_a;
	uint8_t  ident_b;
    } GroupAddress;
    typedef struct
    {
	uint8_t  type;
	uint8_t  pfix;
	uint8_t  ident_a;
	uint8_t  ident_b;
	uint8_t  individual_base_a;
	uint8_t  individual_base_b;
	uint8_t  highest_ident_a;
	uint8_t  highest_ident_b;
	uint8_t  group_base_a;
	uint8_t  group_base_b;
	uint8_t  highest_group_a;
	uint8_t  highest_group_b;
	// Followed by 0 or more GroupAddress
	GroupAddress group[0];
    } MsgNumberingInfo;

    // 5.2.2.8.7
    typedef struct
    {
	uint8_t  type;
	uint8_t  controls_a;
	uint8_t  controls_b;
	uint8_t  controls_c;
	uint8_t  codings;
    } MsgRadioSettings;

    // 5.2.2.8.5
    typedef struct
    {
	uint8_t  type;
	uint8_t  conditions;
	uint8_t  field_strength;
	uint8_t  max_call_duration;
    } MsgOperatingCondition;

    // 5.2.2.8.7
    typedef struct
    {
	uint8_t  type;
	uint8_t  channel_a;
	uint8_t  channel_b;
	uint8_t  sys_ident_a;
	uint8_t  sys_ident_b;
    } MsgNetworkInformation;

#pragma pack()


    Map27DataLayer* _dataLayer;
    Map27ApplicationLayer* _applicationLayer;
    NetworkState _network_state;
    Map27Timer   _request_timer;
};

#endif
