// PhysicalLayer.cpp
//
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2013 Mike McCauley
// $Id: PhysicalLayer.cpp,v 1.8 2013/07/13 11:28:33 mikem Exp $

// For NULL:
#include <string.h>

#include "PhysicalLayer.h"
#include "DataLayer.h"
#include "Port.h"

Map27PhysicalLayer::Map27PhysicalLayer()
{
    _port = NULL;
    _rxState = RX_STATE_IDLE;
}

void Map27PhysicalLayer::reset()
{
    if (_port)
    {
	_port->close();
	_port->open();
    }
    resetRx();
}

void Map27PhysicalLayer::setPort(Map27Port* port)
{
    _port = port;
}

Map27Port* Map27PhysicalLayer::port()
{
    return _port;
}

bool Map27PhysicalLayer::configure_rs232_port(uint32_t baud)
{
    if (_port)
	_port->configure(baud);
}

void Map27PhysicalLayer::setDataLayer(Map27DataLayer* dataLayer)
{
    _dataLayer = dataLayer;
}

Map27DataLayer* Map27PhysicalLayer::dataLayer()
{
    return _dataLayer;
}

void Map27PhysicalLayer::poll()
{
    if (_port)
    {
	while (_port->canRead())
	    handleRx(_port->read());
    }
}

void Map27PhysicalLayer::sendFrameStart()
{
    _txFcs.init();
    // Start of frame
    sendToPort(SYN); // Not in FCS
    sendToPort(DLE); // Not in FCS
    sendToPort(STX); // Not in FCS
}

void Map27PhysicalLayer::sendFrameData(uint8_t *buf, uint16_t len)
{
    // Data with DLE stuffing
    while (len--)
    {
	if (*buf == DLE)    // DLE stuffing required?
	    sendToPort(DLE); // Not in FCS
	_txFcs.update(*buf);
	sendToPort(*buf++);
    }
}

void Map27PhysicalLayer::sendFrameEnd()
{
    // End of frame
    _txFcs.update(DLE);
    sendToPort(DLE);
    _txFcs.update(ETX);
    sendToPort(ETX);
  
    // Now get and send the FCS
    uint16_t fcs = _txFcs.final();
    sendToPort(fcs >> 8);
    sendToPort(fcs & 0xff);
}

void Map27PhysicalLayer::send(uint8_t *buf, uint16_t len)
{
    // This does framing, FCS and DLE stuffing
    sendFrameStart();
    sendFrameData(buf, len);
    sendFrameEnd();
}

void Map27PhysicalLayer::handleRx(uint8_t ch)
{
    // State machine for receiving chars
    switch(_rxState)
    {
	case RX_STATE_IDLE:
	{
	    if (ch == SYN)
		_rxState = RX_STATE_SYN;
	}
	break;

	case RX_STATE_SYN:
	{
	    if (ch == DLE)
		_rxState = RX_STATE_DLE;
	    else if (ch == SYN)
		_rxState = RX_STATE_SYN;
	    else
		_rxState = RX_STATE_IDLE;
	}
	break;

	case RX_STATE_DLE:
	{
	    if (ch == STX)
	    {
		clearRxBuf();
		_rxState = RX_STATE_DATA;
	    }
	    else
		_rxState = RX_STATE_IDLE;
	}
	break;

	case RX_STATE_DATA:
	{
	    if (ch == DLE)
		_rxState = RX_STATE_ESCAPE; // Start escape
	    else
	    {
		if (!appendRxBuffer(ch))
		    _rxState = RX_STATE_IDLE; // Buf Overflow
	    }
	}
	break;

	case RX_STATE_ESCAPE:
	{
	    if (ch == ETX)
	    {
		_rxFcs.update(DLE);
		_rxFcs.update(ch);
		_rxState = RX_STATE_WAIT_FCS1; // End frame
	    }
	    else if (ch == DLE)
	    {
		if (appendRxBuffer(ch))
		    _rxState = RX_STATE_DATA;
		else
		    _rxState = RX_STATE_IDLE; // Buf Overflow
	    }
	    else
		_rxState = RX_STATE_IDLE; // Unexpected
	}
	break;

	case RX_STATE_WAIT_FCS1:
	{
	    _rxRecdFcs = ch << 8;
	    _rxState = RX_STATE_WAIT_FCS2;
	}
	break;

	case RX_STATE_WAIT_FCS2:
	{
	    _rxRecdFcs |= ch;
	    // Check FCS
	    if (_rxRecdFcs == _rxFcs.final())
	    {
		// Process received message
		handleMessage();
	    }
	    else
	    {
		// FCS error
		// REVISIT: count and report errors?
	    }
	    _rxState = RX_STATE_IDLE;
	}
	break;

    }
}

void Map27PhysicalLayer::sendToPort(uint8_t ch)
{
    if (_port)
	_port->send(ch);
}

void Map27PhysicalLayer::sendLinkRequest(uint8_t n1, uint8_t k, uint8_t version)
{
    Map27PhysicalLR lr;

    lr.type = MAP27_PHYSICAL_TYPE_LR;
    lr.n1 = n1;
    lr.k = k;
    lr.version = version;
    send((uint8_t*)&lr, sizeof(lr));
}

void Map27PhysicalLayer::sendLinkAcknowledge(uint8_t n_r, uint8_t n_k)
{
    Map27PhysicalLA la;

    la.type = MAP27_PHYSICAL_TYPE_LA;
    la.n_r = n_r;
    la.n_k = n_k;
    la.reserved = 0;
    send((uint8_t*)&la, sizeof(la));
}

void Map27PhysicalLayer::sendLinkTransfer(uint8_t n_s, uint8_t ar, uint8_t* buf, uint16_t bufLen)
{
    Map27PhysicalLT lt;

    lt.type = MAP27_PHYSICAL_TYPE_LT;
    lt.n_s = n_s;
    lt.ar = ar;
    lt.reserved = 0;
    // Send in 2 lumps to avoid the need for a big buffer
    sendFrameStart();
    sendFrameData((uint8_t*)&lt, sizeof(lt));
    sendFrameData(buf, bufLen);
    sendFrameEnd();
}

void Map27PhysicalLayer::clearRxBuf()
{
    _rxFcs.init();
    _rxBufLen = 0;
}

void Map27PhysicalLayer::handleMessage()
{
    if (_rxBuf[0] == MAP27_PHYSICAL_TYPE_LR
	&& _rxBufLen == sizeof(Map27PhysicalLR))
    {
	Map27PhysicalLR* p = (Map27PhysicalLR*)_rxBuf;
	handleLinkRequest(p->n1, p->k, p->version);
    }
    else if (_rxBuf[0] == MAP27_PHYSICAL_TYPE_LA
	     && _rxBufLen == sizeof(Map27PhysicalLA))
    {
	Map27PhysicalLA* p = (Map27PhysicalLA*)_rxBuf;
	handleLinkAcknowledge(p->n_r, p->n_k);
    }
    else if (_rxBuf[0] == MAP27_PHYSICAL_TYPE_LT
	     && _rxBufLen >= sizeof(Map27PhysicalLT))
    {
	Map27PhysicalLT* p = (Map27PhysicalLT*)_rxBuf;
	handleLinkTransfer(p->n_s, p->ar, p->data, _rxBufLen - sizeof(Map27PhysicalLT));
    }
}

bool Map27PhysicalLayer::appendRxBuffer(uint8_t ch)
{
    if (_rxBufLen < MAP27_MAX_MESSAGE_LEN)
    {
	// Normal data, save and add to FCS
	_rxBuf[_rxBufLen++] = ch;
	_rxFcs.update(ch);
	_rxState = RX_STATE_DATA;
	return true;
    }
    else
    {
	// Buffer overflow
	return false;
    }
}

void Map27PhysicalLayer::handleLinkRequest(uint8_t n1, uint8_t k, uint8_t version)
{
    if (_dataLayer)
	_dataLayer->link_request(n1, k, version);
}

void Map27PhysicalLayer::handleLinkAcknowledge(uint8_t n_r, uint8_t n_k)
{
    if (_dataLayer)
	_dataLayer->link_acknowledge(n_r, n_k);
}

void Map27PhysicalLayer::handleLinkTransfer(uint8_t n_s, uint8_t ar, uint8_t* buf, uint16_t bufLen)
{
    if (_dataLayer)
	_dataLayer->link_transfer(n_s, ar, buf, bufLen);
}

void Map27PhysicalLayer::resetRx()
{
    _rxState = RX_STATE_IDLE;
    _rxBufLen = 0;
}
