// PhysicalLayer.h
//
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2013 Mike McCauley
// $Id: PhysicalLayer.h,v 1.5 2013/07/13 11:28:33 mikem Exp $


#ifndef MAP27_PHYSICALLAYER_H
#define MAP27_PHYSICALLAYER_H

#include <stdint.h>

#include "Map27.h"
#include "FCS.h"

class Map27Port;
class Map27DataLayer;

// Special characters
#define STX 0x02
#define ETX 0x03
#define DLE 0x10
#define SYN 0x16

// Physical Layer Message types
#define MAP27_PHYSICAL_TYPE_LR              0x01
#define MAP27_PHYSICAL_TYPE_LA              0x02
#define MAP27_PHYSICAL_TYPE_LT              0x04

/////////////////////////////////////////////////////////////////////
/// \class Map27PhysicalLayer PhysicalLayer.h <Map27/PhysicalLayer.h>
/// \brief Manage the Physical Layer of the Map27 protocol stack
///
/// This class reads data from the device Port and formats the data into complete frames
/// and checks their FCS.
/// Its also formats outdoing messages with framing and frame check characters
class Map27PhysicalLayer
{
public:
    /// Constructor
    Map27PhysicalLayer();

    /// Sets the pointer to the Port to use
    /// to communicate with the Radio
    /// \param[in] port Pointer to the Port
    void setPort(Map27Port* port);

    /// Get the current Port pointer
    /// \return The current Port pointer, or NULL if previously set.
    Map27Port* port();

    /// Sets the pointer to the Data LAyer to use
    /// to deliver messages up the protocol stack
    /// \param[in] dataLayer Pointer to the Data Layer
    void setDataLayer(Map27DataLayer* dataLayer);

    /// Get the current Data Layer pointer
    /// \return The current Data Layer pointer, or NULL if previously set.
    Map27DataLayer* dataLayer();

    /// Resets the port
    /// by closing and reopenng it
    void reset();

    /// Configures the desiredbaud
    /// rate to use on the Port the next time it is opened
    /// \param[in] baud The baud rate to set, in bauds (per second)
    /// \returns true if successful
    bool configure_rs232_port(uint32_t baud = 9600);

    /// Sends a complete message to the radio.
    /// Prefixes the message withthe frame start, does data stuffing, and 
    /// adds the frame end and FCS
    /// \param[in] buf Pointer to the data payload to send
    /// \param[in] len Number of bytes in buf
    void send(uint8_t *buf, uint16_t len);

    /// Sends a single octet to
    /// the current Port
    /// \param[in] ch Theoctet to send
    virtual void sendToPort(uint8_t ch);

    /// Sends a Link Request (LR) message
    /// to the Port
    /// \param[in] n1 The requested maimum message size
    /// \param[in] k The requested maximum window size
    /// \param[in] version The requested protocol version
    void sendLinkRequest(uint8_t n1, uint8_t k, uint8_t version);

    /// Sends a Link Acknowledge (LA) messgae
    /// to the Port
    /// \param[in] n_r receive sequence number
    /// \param[in] n_k Receive credit number
    void sendLinkAcknowledge(uint8_t n_r, uint8_t n_k);

    /// Sends a Link Transfer (LT) messgae
    /// to the Port
    /// \param[in] n_s Transmit sequence number
    /// \param[in] ar Immediate Acknowledgement request
    /// \param[in] buf Pointer to the payload data
    /// \param[in] bufLen Number of bytes in buf
    void sendLinkTransfer(uint8_t n_s, uint8_t ar, uint8_t* buf, uint16_t bufLen);

    /// Poll the Physical Layer for timer and input events from lower layers in the protocol stack
    /// In Unix/Linux and Arduino environments, you are required to call this at least as 
    /// frequently as octets might be received from the radio. At least every 10msec would 
    /// be appropriate for a 9600 baud connection.
    /// Note: Map27DataLayer does this automatically when its poll() function is called.
    void poll();

    /// Clears the incoming
    /// message buffer, perhaps to due a link reset
    void clearRxBuf();

private:
    void handleRx(uint8_t ch);
    virtual void handleMessage();
    bool appendRxBuffer(uint8_t ch);
    virtual void handleLinkRequest(uint8_t n1, uint8_t k, uint8_t version);
    virtual void handleLinkAcknowledge(uint8_t n_r, uint8_t n_k);
    virtual void handleLinkTransfer(uint8_t n_s, uint8_t ar, uint8_t* buf, uint16_t bufLen);
    void sendFrameStart();
    void sendFrameData(uint8_t *buf, uint16_t len);
    void sendFrameEnd();
    void resetRx();

    // Receiver states
    typedef enum
    {
	RX_STATE_IDLE      = 1,
	RX_STATE_SYN       = 2,
	RX_STATE_DLE       = 3,
	RX_STATE_DATA      = 4,
	RX_STATE_ESCAPE    = 5,
	RX_STATE_WAIT_FCS1 = 6,
	RX_STATE_WAIT_FCS2 = 7
    } PhysicalRxState;

    // Physical Layer header structures
    typedef struct
    {
	uint8_t     type;     // MAP27_PHYSICAL_TYPE_LR
	uint8_t     n1;       // Maximum length
	uint8_t     k;        // Window size
	uint8_t     version;  // MAP27_PHYSICAL_VERSION
    } Map27PhysicalLR;
  
    typedef struct
    {
	uint8_t     type;     // MAP27_PHYSICAL_TYPE_LA
	uint8_t     n_r;      // Rx sequence number N(R)
	uint8_t     n_k;      // Rx credit number N(k)
	uint8_t     reserved; // 0
    } Map27PhysicalLA;
  
    typedef struct
    {
	uint8_t     type;     // MAP27_PHYSICAL_TYPE_LT
	uint8_t     n_s;      // Tx sequence number
	uint8_t     ar;       // Acknowledgement request 0/1
	uint8_t     reserved; // 0
	uint8_t     data[0];
    } Map27PhysicalLT;
  
    Map27Port* _port;
    Map27DataLayer* _dataLayer;
    Map27FCS   _rxFcs;
    Map27FCS   _txFcs;
    PhysicalRxState    _rxState;
    uint8_t    _rxBuf[MAP27_MAX_MESSAGE_LEN];
    uint16_t   _rxBufLen;
    uint16_t   _rxRecdFcs;
};

#endif
