// Port.h
//
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2013 Mike McCauley
// $Id: Port.h,v 1.5 2013/07/09 23:35:55 mikem Exp $

#ifndef MAP27_PORT_H
#define MAP27_PORT_H

#include <stdint.h>

/////////////////////////////////////////////////////////////////////
/// \class Map27Port Port.h <Map27/Port.h>
/// \brief Abstract superclass for accessing RS232 port on the host
///
/// This class (or subclasses) are used by Map27PhysicalLayer to send and receive data to and 
/// from an RS232 port on the host computer.
///
/// At present the only concrete implementations are:
/// - Map27UnixPort Access to Unix/Linux RS232 serial ports using termio
/// The most appropriate implementaiton for your OS wil be selecterd by Port.cpp, so you only
/// need tolink to Port.o (or its equivalent for your OS).
class Map27Port
{
public:
    /// Constructor
    Map27Port();

    /// Sets the baud rate to use.
    /// Map27 mandates 8N1 format, but the baud rate is expected to be configurable to at least 1200 and 9600 baud
    /// This interface allows any supported baud rate to be set.
    /// \param[in] baud The desired baud rate in bauds (ie per-second)
    virtual bool configure(uint32_t baud = 9600);

    /// Open the underlying physical device
    /// Sets the baud rate to the configured rate and to 8N1
    virtual bool open();

    /// Close the underlying physical device
    /// if it is currently open
    virtual bool close();

    /// Sends a single octet to the underlying physical device
    /// \param[in] ch The character to send.
    /// Blocks until the character is queued by the operating system
    /// \param[in] ch The octet to send
    virtual void send(uint8_t ch);

    /// Test whether readable input characters are waiting
    /// \return true if one or more octets are waiting to be read from the underlying device
    virtual bool canRead();

    /// Reads a single octet from the underlying device
    /// \return The next octet to be read. Undefined if canRead() is not true.
    virtual uint8_t read();

protected:
    /// The currently configured baud rate
    uint32_t _baud;
};


#endif
