// Timer.h

#ifndef MAP27_TIMER_H
#define MAP27_TIMER_H

#include <stdint.h>
#include <sys/time.h>

/////////////////////////////////////////////////////////////////////
/// \class Map27Timer Timer.h <Map27/Timer.h>
/// \brief Class for one-shot millisecond timers.
///
/// This class implements a basic one-shot timer. 
/// The current implementation is intended to be polled periodically
/// to determine whether the set time has expire, although this can be changed
/// to suit event-driven environments.
/// The current implementation uses Unix based times, but can be ported to other architectures.
/// 
/// Each timer has a time interval in milliseconds, and callback which will be called when the timer expires
/// (or at least it will be called from within poll() the first time poll() is called after the time would expire.
///
/// The usual mode of use is:
/// \code
/// #define INTERVAL 12345
/// Timer mytimer;
/// mytimer.setCallback(mycallbackfunction, mycallbackarg);
/// mytimer.start(INTERVAL);
/// mytimer.poll();
/// ....
/// mytimer.poll();
/// .... // do some things, during which time the INTERVAL expires
/// ///(after INTERVAL has expired):
/// mytimer.poll();
/// // poll() calls mycallbackfunction(mycallbackarg);
/// // timer is now inactive until start() is called again
///\endcode
class Map27Timer
{
public:

    /// Constructor
    /// Defaults to inactive, with an interval of 0 msec.
    Map27Timer();

    /// Stop the timer
    /// The timer econes inactive and will not trigger
    void stop();

    /// Start (or restart) the timer, using the last set interval.
    /// Any unexpired potion of the current timer is lost, and the interval
    /// starts again from now.
    void start();

    /// Start (or restart) the timer, using the given interval.
    /// Any unexpired potion of the current timer is lost, and the interval
    /// starts again from now.
    /// \param[in] interval The timer interval in milliseconds.
    void start(uint32_t interval);

    /// Poll the timer for expiry.
    /// If more than interval milliconds have past sine the last start, trigger the callback
    /// and mark the timer as inactive (ie this as a one-shot timer)
    void poll();

    /// Set the desired interval.
    /// Any previously set interval is lost. If the new interval is 
    /// less than the time since the last start() the timer will trigger.
    /// \param[in] interval The timer interval in in illisconds.
    void setInterval(uint32_t interval);

    /// Set the callback function.
    /// The callback function will be called whenever the timer triggers.
    /// When the timer triggers, the calback function wil be called like:
    /// \code
    /// callback(callbackArg);
    /// \endcode
    /// \param[in] callback Pointer to the allback function
    /// \param[in] callbackArg OPtional argument to be p[assed to callback()
    void setCallback(void (*callback)(void*), void* callbackArg = 0);

    /// Get the current timer interval
    /// \return The current timer interval
    uint32_t interval();

    /// Determine if the timer is active
    /// \return true if the timer is currently active.
    bool active();

protected:
    /// Called when the timer is triggered.
    void triggered();

private:
    uint32_t       _interval; // Milliseconds
    bool           _active;   // true if the timer is active
    struct timeval _startTime; // Timer of the last start()
    void           (*_callback)(void*); // Callback function
    void*          _callbackArg; // Callback argument
};

#endif
