// UnixPort.h
//
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2013 Mike McCauley
// $Id: UnixPort.cpp,v 1.7 2013/07/13 11:28:33 mikem Exp $

#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>

#include "UnixPort.h"
#include "Log.h"

Map27UnixPort::Map27UnixPort()
{
    _fd = 0;
    _isOpen = false;
    _device = "/dev/ttyUSB0";
}

void Map27UnixPort::setDevice(const char* device)
{
    _device = device;
}

const char* Map27UnixPort::device()
{
    return _device;
}

// Map numeric baud rates to TTY port constants
static speed_t get_baud_constant(uint32_t baud)
{
    switch (baud)
    {    
	case 50:
	    return B50;
	case 75:
	    return B75;
	case 110:
	    return B110;
	case 134:
	    return B134;
	case 150:
	    return B150;
	case 200:
	    return B200;
	case 300:
	    return B300;
	case 600:
	    return B600;
	case 1200:
	    return B1200;
	case 1800:
	    return B1800;
	case 2400:
	    return B2400;
	case 4800:
	    return B4800;
	case 9600:
	    return B9600;
	case 19200:
	    return B19200;
	case 38400:
	    return B38400;
	case 57600:
	    return B57600;
	case 115200:
	    return B115200;
	case 230400:
	    return B230400;
	default:
	    fprintf(stderr, "UnixPort: invalid baud rate: %d. Using 9600\n", baud);
	    return B9600;

    }
}

bool Map27UnixPort::open()
{
    if ((_fd = ::open(_device, O_RDWR | O_NOCTTY | O_NDELAY)) != -1) 
    {        
	::tcgetattr(_fd, &_oldTermios);
        _currentTermios = _oldTermios;
        ::cfmakeraw(&_currentTermios);

	// Configured baud
	speed_t baud = get_baud_constant(_baud);
	::cfsetispeed(&_currentTermios, baud);
	::cfsetospeed(&_currentTermios, baud);

	// Map27 is always 8N1
	_currentTermios.c_cflag |= CS8;       // 8 bits
	_currentTermios.c_cflag &= (~PARENB); // None
	_currentTermios.c_cflag &= (~CSTOPB); // 1 stop bit

	::tcsetattr(_fd, TCSAFLUSH, &_currentTermios);
	_isOpen = true;
	return true;
    }
    else
    {
	_isOpen = false;
	fprintf(stderr, "Map27UnixPort open %s failed: %s\n", _device, ::strerror(errno));
	return false;
    }
}

bool Map27UnixPort::close()
{
    if (_isOpen)
    {
	::tcsetattr(_fd, TCSAFLUSH | TCSANOW, &_oldTermios);   // Restore termios
	::close(_fd);
	_isOpen = false;
    }
    return true;
}

void Map27UnixPort::send(uint8_t ch)
{
//    DEBUG_PRINTF("port send %02x\n", ch);
    // Note: blocks until completion
    ::write(_fd, &ch, 1);
}

bool Map27UnixPort::canRead()
{    
    int bytesQueued = 0;

    ::ioctl(_fd, FIONREAD, &bytesQueued);
//    DEBUG_PRINTF("port canRead %d\n", bytesQueued);
    return (bytesQueued > 0);
}

uint8_t Map27UnixPort::read()
{
    uint8_t data;

    ::read(_fd, &data, 1);
//    DEBUG_PRINTF("port read %02x\n", data);
    return data;
}
