// UnixPort.h
//
// Author: Mike McCauley (mikem@airspayce.com)
// Copyright (C) 2013 Mike McCauley
// $Id: UnixPort.h,v 1.4 2013/07/13 11:28:33 mikem Exp $

#ifndef MAP27_UNIXPORT_H
#define MAP27_UNIXPORT_H

#include <stdint.h>
#include <termios.h>

#include"Port.h"

/////////////////////////////////////////////////////////////////////
/// \class Map27UnixPort UnixPort.h <Map27/UnixPort.h>
/// \brief Access RS232 port on Unix/Linux
///
/// This is a concrete implementation of Map27Port for use on Unix/Linux platforms.
/// It is automatically selected by Port.cpp at compile time, so you should only need to link
/// with Port.o.
///
/// You should specify the underlying device to use with setDevice(), which defaults to '/dev/ttyUSB0'.
class Map27UnixPort : public Map27Port
{
public:

    /// Constructor
    /// The default device is '/dev/ttyUSB0'.
    Map27UnixPort();

    /// Sets the name of the underlying RS232 device to open.
    /// Defaults to '/dev/ttyUSB0'.
    /// \param[in] device ASCII name of the UNix device of the RS232 port to open
    void setDevice(const char* device);

    /// Get the currently configured device
    /// \return The currently configured RS232 device name
    const char* device();

    /// Open the underlying physical device
    /// Sets the baud rate to the configured rate and to 8N1
    virtual bool open();

    /// Close the underlying physical device
    /// if it is currently open
    virtual bool close();

    /// Sends a single octet to the underlying physical device
    /// \param[in] ch The character to send.
    /// Blocks until the character is queued by the operating system
    /// \param[in] ch The octet to send
    virtual void send(uint8_t ch);

    /// Test whether readable input characters are waiting
    /// \return true if one or more octets are waiting to be read from the underlying device
    virtual bool canRead();

    /// Reads a single octet from the underlying device
    /// \return The next octet to be read. Undefined if canRead() is not true.
    virtual uint8_t read();

private:
    const char* _device;  // ASCII device name
    int _fd;              // filedesc if open
    bool _isOpen ;        // true if device is open
    struct termios _currentTermios; // Current terminal settings
    struct termios _oldTermios; // Original terminal settings
};

#endif
